"""
Tests for :py:mod:`line_profiler.autoprofile.eager_preimports`.

Notes
-----
Most of the features are already covered by the doctests.
"""
import subprocess
import sys
from contextlib import ExitStack
from functools import partial
from pathlib import Path
from operator import methodcaller
from runpy import run_path
from tempfile import TemporaryDirectory
from textwrap import dedent
from types import SimpleNamespace
from typing import Collection, Generator, Sequence, Type, Optional, Union
from uuid import uuid4
from warnings import catch_warnings

import pytest
try:
    import flake8  # noqa: F401
except ImportError:
    HAS_FLAKE8 = False
else:
    HAS_FLAKE8 = True

from line_profiler.autoprofile.eager_preimports import (
    split_dotted_path, resolve_profiling_targets, write_eager_import_module)


def write(path: Path, content: Optional[str] = None) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    if content is None:
        path.touch()
    else:
        path.write_text(dedent(content).strip('\n'))


def gen_names(name) -> Generator[str, None, None]:
    while True:
        yield '_'.join([name, *str(uuid4()).split('-')])


@pytest.fixture
def preserve_sys_state() -> None:
    old_path = sys.path.copy()
    old_modules = sys.modules.copy()
    try:
        yield
    finally:
        sys.path.clear()
        sys.path[:] = old_path
        sys.modules.clear()
        sys.modules.update(old_modules)


@pytest.fixture
def sample_package(preserve_sys_state: None, tmp_path: Path) -> str:
    """
    Write a normal package and put it in :py:data:`sys.path`.  When
    we're done, reset :py:data:`sys.path` and `sys.modules`.
    """
    module_name = next(name for name in gen_names('my_sample_pkg')
                       if name not in sys.modules)
    new_path = tmp_path / '_modules'
    write(new_path / module_name / '__init__.py')
    write(new_path / module_name / 'foo' / '__init__.py')
    write(new_path / module_name / 'foo' / 'bar.py')
    write(new_path / module_name / 'foo' / 'baz.py',
          """
          '''
          This is a bad module.
          '''
          raise AssertionError
          """)
    write(new_path / module_name / 'foobar.py')
    # Cleanup managed with `preserve_sys_state()`
    sys.path.insert(0, str(new_path))
    yield module_name


@pytest.fixture
def sample_namespace_package(
        preserve_sys_state: None,
        tmp_path_factory: pytest.TempPathFactory) -> str:
    """
    Write a namespace package and put it in :py:data:`sys.path`.  When
    we're done, reset :py:data:`sys.path` and `sys.modules`.
    """
    module_name = next(name for name in gen_names('my_sample_namespace_pkg')
                       if name not in sys.modules)
    new_paths = [tmp_path_factory.mktemp('_modules-', numbered=True)
                 for _ in range(3)]
    for submod, new_path in zip(['one', 'two', 'three'], new_paths):
        write(new_path / module_name / (submod + '.py'))
        # Cleanup managed with `preserve_sys_state()`
        sys.path.insert(0, str(new_path))
    yield module_name


@pytest.mark.parametrize(
    ('adder', 'xc'),
    [('foo; bar', ValueError), (1, TypeError), ('(foo\n .bar)', ValueError)])
def test_write_eager_import_module_wrong_adder(
        adder: str, xc: Type[Exception]) -> None:
    """
    Test passing an erroneous ``adder`` to
    :py:meth:`~.write_eager_import_module()`.
    """
    with pytest.raises(xc):
        write_eager_import_module(['foo'], adder=adder)


@pytest.mark.skipif(not HAS_FLAKE8, reason='no `flake8`')
def test_written_module_pep8_compliance(sample_package: str):
    """
    Test that the module written by
    :py:meth:`~.write_eager_import_module()` passes linting by
    :py:mod:`flake8`.
    """
    with TemporaryDirectory() as tmpdir:
        module = Path(tmpdir) / 'module.py'
        with module.open(mode='w') as fobj:
            write_eager_import_module(
                [sample_package + '.foobar'],
                recurse=[sample_package + '.foo'], stream=fobj)
        print(module.read_text())
        (subprocess
         .run([sys.executable, '-m', 'flake8',
               '--extend-ignore=E501',  # Allow long lines
               module])
         .check_returncode())


@pytest.mark.parametrize(
    ('dotted_paths', 'recurse', 'warnings', 'error'),
    [(['__MODULE__.foobar'], ['__MODULE__.foo'],
      # `foo.baz` is indirectly included, so its raising an error
      # shouldn't cause the script to error out
      [{'target cannot', '__MODULE__.foo.baz'}],
      None),
     # We don't recurse down `__MODULE__.foo`, so that doesn't give a
     # warning; but `__MODULE__.baz` cannot be imported because it
     # doesn't exist
     (['__MODULE__.foo', '__MODULE__.baz'], False,
      [{'target cannot', '__MODULE__.baz'}], None),
     # If we do recurse however, `__MODULE__.foo.baz` also ends up in
     # the warning
     # (also there's a `__MODULE___foo` which doesn't exist, about which
     # the warning is issued during module generation)
     (['__MODULE__' + '_foo', '__MODULE__', '__MODULE__.baz'], True,
      [{'target cannot', '__MODULE__' + '_foo'},  # Fails at write
       {'targets cannot',  # Fails at import
        '__MODULE__.foo.baz', '__MODULE__.baz'}],
      None),
     # And if the problematic module is an explicit target, raise the
     # error
     (['__MODULE__', '__MODULE__.foo.baz'], False, [], AssertionError)])
def test_written_module_error_handling(
        sample_package: str,
        dotted_paths: Collection[str],
        recurse: Union[Collection[str], bool],
        warnings: Sequence[Collection[str]],
        error: Union[Type[Exception], None]):
    """
    Test that the module written by
    :py:meth:`~.write_eager_import_module()` gracefully handles errors
    for implicitly included modules.
    """
    replace = methodcaller('replace', '__MODULE__', sample_package)
    dotted_paths = [replace(target) for target in dotted_paths]
    if recurse not in (True, False):
        recurse = [replace(target) for target in recurse]
    warnings = [{replace(fragment) for fragment in fragments}
                for fragments in warnings]
    with TemporaryDirectory() as tmpdir:
        module = Path(tmpdir) / 'module.py'
        with ExitStack() as stack:
            enter = stack.enter_context
            # Set up the warning capturing early so that we catch both
            # warnings at module-generation time and execution time
            captured_warnings = enter(catch_warnings(record=True))
            with module.open(mode='w') as fobj:
                write_eager_import_module(
                    dotted_paths, recurse=recurse, stream=fobj)
            print(module.read_text())
            if error is not None:
                enter(pytest.raises(error))
            # Just use a dummy object, no need to instantiate a profiler
            prof = SimpleNamespace(
                add_imported_function_or_module=lambda *_, **__: 0)
            run_path(str(module), {'profile': prof}, 'module')
    assert len(captured_warnings) == len(warnings)
    for warning, fragments in zip(captured_warnings, warnings):
        for fragment in fragments:
            assert fragment in str(warning.message)


def test_split_dotted_path_staticity() -> None:
    """
    Test `split_dotted_path()` with different values for `static`.
    """
    split = partial(split_dotted_path, 'os.path.abspath')
    # Static analysis has no idea of `os.path` members since `os.path`
    # is dynamically imported from e.g. `posixpath` or `ntpath`
    assert split(static=True) == ('os', 'path.abspath')
    # The import system knows of the already-imported module `os.path`
    assert split(static=False) == ('os.path', 'abspath')


def test_resolve_profiling_targets_staticity(
        sample_namespace_package: str) -> None:
    """
    Test subpackage/-module discovery with `resolve_profiling_targets()`
    with different values for `static`.
    """
    all_targets = ({f'{sample_namespace_package}.{submod}'
                    for submod in ['one', 'two', 'three']}
                   | {sample_namespace_package})
    # Static analysis can't handle namespace packages
    resolve = partial(
        resolve_profiling_targets, [sample_namespace_package], recurse=True)
    static_result = resolve(static=True)
    assert set(static_result.targets) < all_targets, static_result
    # The import system successfully retrieves all submodules
    dyn_result = resolve(static=False)
    assert set(dyn_result.targets) == all_targets, dyn_result



"""
Miscellaneous utilities that :py:mod:`line_profiler` uses.
"""
import enum


class _StrEnumBase(str, enum.Enum):
    """
    Base class mimicking :py:class:`enum.StrEnum` in Python 3.11+.

    Example
    -------
    >>> import enum
    >>>
    >>>
    >>> class MyEnum(_StrEnumBase):
    ...     foo = enum.auto()
    ...     BAR = enum.auto()
    ...
    >>>
    >>> MyEnum.foo
    <MyEnum.foo: 'foo'>
    >>> MyEnum('bar')
    <MyEnum.BAR: 'bar'>
    >>> MyEnum('baz')
    Traceback (most recent call last):
      ...
    ValueError: 'baz' is not a valid MyEnum
    """
    @staticmethod
    def _generate_next_value_(name, *_, **__):
        return name.lower()

    def __eq__(self, other):
        return self.value == other

    def __str__(self):
        return self.value


class StringEnum(getattr(enum, 'StrEnum', _StrEnumBase)):
    """
    Convenience wrapper around :py:class:`enum.StrEnum`.

    Example
    -------
    >>> import enum
    >>>
    >>>
    >>> class MyEnum(StringEnum):
    ...     foo = enum.auto()
    ...     BAR = enum.auto()
    ...
    >>>
    >>> MyEnum.foo
    <MyEnum.foo: 'foo'>
    >>> MyEnum('bar')
    <MyEnum.BAR: 'bar'>
    >>> bar = MyEnum('BAR')  # Case-insensitive
    >>> bar
    <MyEnum.BAR: 'bar'>
    >>> assert isinstance(bar, str)
    >>> assert bar == 'bar'
    >>> str(bar)
    'bar'
    """
    @classmethod
    def _missing_(cls, value):
        if not isinstance(value, str):
            return None
        members = {name.casefold(): instance
                   for name, instance in cls.__members__.items()}
        return members.get(value.casefold())



from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import Mapping, Sequence, Any, Self, TypeVar


TARGETS = 'line_profiler.toml', 'pyproject.toml'
ENV_VAR = 'LINE_PROFILER_RC'

K = TypeVar('K')
V = TypeVar('V')
Config = tuple[dict[str, dict[str, Any]], Path]
NestedTable = Mapping[K, 'NestedTable[K, V]' | V]


@dataclass
class ConfigSource:
    conf_dict: dict[str, Any]
    path: Path
    subtable: list[str]

    def copy(self) -> Self:
        ...

    def get_subconfig(self, *headers: str,
                      allow_absence: bool = False, copy: bool = False) -> Self:
        ...

    @classmethod
    def from_default(cls, *, copy: bool = True) -> Self:
        ...

    @classmethod
    def from_config(cls, config: str | PathLike | bool | None = None, *,
                    read_env: bool = True) -> Self:
        ...


def find_and_read_config_file(
        *,
        config: str | PathLike | None = None,
        env_var: str | None = ENV_VAR,
        targets: Sequence[str | PathLike] = TARGETS) -> Config:
    ...


def get_subtable(table: NestedTable[K, V], keys: Sequence[K], *,
                 allow_absence: bool = True) -> NestedTable[K, V]:
    ...


def get_headers(table: NestedTable[K, Any], *,
                include_implied: bool = False) -> set[tuple[K, ...]]:
    ...



import os
import re
import sys
import tempfile
from contextlib import ExitStack

import pytest
import ubelt as ub


class enter_tmpdir:
    """
    Set up a temporary directory and :cmd:`chdir` into it.
    """
    def __init__(self):
        self.stack = ExitStack()

    def __enter__(self):
        """
        Returns:
            curdir (ubelt.Path)
                Temporary directory :cmd:`chdir`-ed into.

        Side effects:
            ``curdir`` created and :cmd:`chdir`-ed into.
        """
        enter = self.stack.enter_context
        tmpdir = os.path.abspath(enter(tempfile.TemporaryDirectory()))
        enter(ub.ChDir(tmpdir))
        return ub.Path(tmpdir)

    def __exit__(self, *_, **__):
        """
        Side effects:
            * Original working directory restored.
            * Temporary directory created deleted.
        """
        self.stack.close()


class restore_sys_modules:
    """
    Restore :py:attr:`sys.modules` after exiting the context.
    """
    def __enter__(self):
        self.old = sys.modules.copy()

    def __exit__(self, *_, **__):
        sys.modules.clear()
        sys.modules.update(self.old)


def write(path, code=None):
    path.parent.mkdir(exist_ok=True, parents=True)
    if code is None:
        path.touch()
    else:
        path.write_text(ub.codeblock(code))


def test_simple_explicit_nonglobal_usage():
    """
    python -c "from test_explicit_profile import *; test_simple_explicit_nonglobal_usage()"
    """
    from line_profiler import LineProfiler
    profiler = LineProfiler()

    def func(a):
        return a + 1

    profiled_func = profiler(func)

    # Run Once
    profiled_func(1)

    lstats = profiler.get_stats()
    print(f'lstats.timings={lstats.timings}')
    print(f'lstats.unit={lstats.unit}')
    print(f'profiler.code_hash_map={profiler.code_hash_map}')
    profiler.print_stats()


def _demo_explicit_profile_script():
    return ub.codeblock(
        '''
        from line_profiler import profile

        @profile
        def fib(n):
            a, b = 0, 1
            for _ in range(n):
                a, b = b, a + b
            return a
        fib(10)
        ''')


def test_explicit_profile_with_nothing():
    """
    Test that no profiling happens when we dont request it.
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())

            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        assert not (temp_dpath / 'profile_output.txt').exists()
        assert not (temp_dpath / 'profile_output.lprof').exists()


def test_explicit_profile_with_environ_on():
    """
    Test that explicit profiling is enabled when we specify the LINE_PROFILE
    enviornment variable.
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        env = os.environ.copy()
        env['LINE_PROFILE'] = '1'

        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())

            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args, env=env)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        assert (temp_dpath / 'profile_output.txt').exists()
        assert (temp_dpath / 'profile_output.lprof').exists()


def test_explicit_profile_ignores_inherited_owner_marker():
    """
    Standalone runs should not be blocked by an inherited owner marker.
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        env = os.environ.copy()
        env['LINE_PROFILE'] = '1'
        env['LINE_PROFILER_OWNER_PID'] = str(os.getpid() + 100000)
        env['PYTHONPATH'] = os.getcwd()

        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())

            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args, env=env)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        assert (temp_dpath / 'profile_output.txt').exists()
        assert (temp_dpath / 'profile_output.lprof').exists()


def test_explicit_profile_process_pool_forkserver():
    """
    Ensure explicit profiler works with forkserver ProcessPoolExecutor.
    """
    import multiprocessing as mp
    if 'forkserver' not in mp.get_all_start_methods():
        pytest.skip('forkserver start method not available')
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        env = os.environ.copy()
        env['LINE_PROFILE'] = '1'
        # env['LINE_PROFILER_DEBUG'] = '1'
        env['PYTHONPATH'] = os.getcwd()

        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(ub.codeblock(
                '''
                import multiprocessing as mp
                from concurrent.futures import ProcessPoolExecutor
                from line_profiler import profile

                def worker(x):
                    return x * x

                @profile
                def run():
                    total = 0
                    for i in range(1000):
                        total += i % 7
                    with ProcessPoolExecutor(max_workers=2) as ex:
                        list(ex.map(worker, range(4)))
                    return total

                def main():
                    if 'forkserver' in mp.get_all_start_methods():
                        mp.set_start_method('forkserver', force=True)
                    run()

                if __name__ == '__main__':
                    main()
                ''').strip())

            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args, env=env)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        output_path = temp_dpath / 'profile_output.txt'
        assert output_path.exists()
        assert output_path.stat().st_size > 100
        assert proc.stdout.count('Wrote profile results to profile_output.txt') == 1


def test_explicit_profile_with_environ_off():
    """
    When LINE_PROFILE is falsy, profiling should not run.
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        env = os.environ.copy()
        env['LINE_PROFILE'] = '0'

        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())

            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        assert not (temp_dpath / 'profile_output.txt').exists()
        assert not (temp_dpath / 'profile_output.lprof').exists()


def test_explicit_profile_with_cmdline():
    """
    Test that explicit profiling is enabled when we specify the --line-profile
    command line flag.

    xdoctest ~/code/line_profiler/tests/test_explicit_profile.py test_explicit_profile_with_environ
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())

            args = [sys.executable, os.fspath(script_fpath), '--line-profile']
            print(f'args={args}')
            proc = ub.cmd(args)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        assert (temp_dpath / 'profile_output.txt').exists()
        assert (temp_dpath / 'profile_output.lprof').exists()


@pytest.mark.parametrize('line_profile', [True, False])
def test_explicit_profile_with_kernprof(line_profile: bool):
    """
    Test that explicit profiling works when using kernprof. In this case
    we should get as many output files.
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)
        base_cmd = [sys.executable, '-m', 'kernprof']
        if line_profile:
            base_cmd.append('-l')
            outfile = 'script.py.lprof'
        else:
            outfile = 'script.py.prof'

        with ub.ChDir(temp_dpath):
            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())
            args = base_cmd + [os.fspath(script_fpath)]
            proc = ub.cmd(args)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        assert not (temp_dpath / 'profile_output.txt').exists()
        assert (temp_dpath / outfile).exists()


@pytest.mark.parametrize('package', [True, False])
@pytest.mark.parametrize('builtin', [True, False])
def test_explicit_profile_with_kernprof_m(builtin: bool, package: bool):
    """
    Test that explicit (non-line) profiling works when using
    `kernprof -m` to run packages and/or submodules with relative
    imports.

    Parameters:
        builtin (bool)
            Whether to slip `@profile` into the globals with `--builtin`
            (true) or to require importing it from `line_profiler` in
            the profiled source code (false)

        package (bool)
            Whether to add the code to a package's `__main__.py` and
            `kernprof -m {<package>}` (true), or to add it to a
            submodule and `kernprof -m {<package>}.{<submodule>}`
            (false)
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)

        lib_code = ub.codeblock(
            '''
            @profile
            def func1(a):
                return a + 1

            @profile
            def func2(a):
                return a + 1

            def func3(a):
                return a + 1

            def func4(a):
                return a + 1
            ''').strip()
        if not builtin:
            lib_code = 'from line_profiler import profile\n' + lib_code
        target_code = ub.codeblock(
            '''
            from ._lib import func1, func2, func3, func4

            if __name__ == '__main__':
                func1(1)
                func2(1)
                func3(1)
                func4(1)
            ''').strip()

        if package:
            target_module = 'package'
            target_fname = '__main__.py'
        else:
            target_module = 'package.api'
            target_fname = 'api.py'

        args = ['kernprof', '-v', '-m', target_module]
        if builtin:
            args.insert(2, '--builtin')  # Insert before the `-m` flag

        if 'PYTHONPATH' in os.environ:
            python_path = '{}:{}'.format(os.environ['PYTHONPATH'], os.curdir)
        else:
            python_path = os.curdir
        env = {**os.environ, 'PYTHONPATH': python_path}

        with ub.ChDir(temp_dpath):
            package_dir = ub.Path('package').mkdir()

            lib_fpath = package_dir / '_lib.py'
            lib_fpath.write_text(lib_code)

            target_fpath = package_dir / target_fname
            target_fpath.write_text(target_code)

            (package_dir / '__init__.py').touch()

            proc = ub.cmd(args, env=env)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        # Note: in non-builtin mode, the entire script is profiled
        for func, profiled in [('func1', True), ('func2', True),
                               ('func3', not builtin), ('func4', not builtin)]:
            result = re.search(r'lib\.py:[0-9]+\({}\)'.format(func),
                               proc.stdout)
            assert bool(result) == profiled

        assert not (temp_dpath / 'profile_output.txt').exists()
        assert (temp_dpath / (target_module + '.prof')).exists()


def test_explicit_profile_with_in_code_enable():
    """
    Test that the user can enable the profiler explicitly from within their
    code.

    CommandLine:
        pytest tests/test_explicit_profile.py -s -k test_explicit_profile_with_in_code_enable
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)

        code = ub.codeblock(
            '''
            from line_profiler import profile
            import ubelt as ub
            print('')
            print('')
            print('start test')

            print('profile = {}'.format(ub.urepr(profile, nl=1)))
            print(f'profile._profile={profile._profile}')
            print(f'profile.enabled={profile.enabled}')

            @profile
            def func1(a):
                return a + 1

            profile.enable(output_prefix='custom_output')

            print('profile = {}'.format(ub.urepr(profile, nl=1)))
            print(f'profile._profile={profile._profile}')
            print(f'profile.enabled={profile.enabled}')

            @profile
            def func2(a):
                return a + 1

            print('func2 = {}'.format(ub.urepr(func2, nl=1)))

            profile.disable()

            @profile
            def func3(a):
                return a + 1

            profile.enable()

            @profile
            def func4(a):
                return a + 1

            func1(1)
            func2(1)
            func3(1)
            func4(1)

            profile._profile
            ''')
        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(code)

            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        print('Finished running script')

        output_fpath = (temp_dpath / 'custom_output.txt')
        raw_output = output_fpath.read_text()
        print(f'Contents of {output_fpath}')
        print(raw_output)

        assert 'Function: func1' not in raw_output
        assert 'Function: func2' in raw_output
        assert 'Function: func3' not in raw_output
        assert 'Function: func4' in raw_output

        assert output_fpath.exists()
        assert (temp_dpath / 'custom_output.lprof').exists()


def test_explicit_profile_with_duplicate_functions():
    """
    Test profiling duplicate functions with the explicit profiler

    CommandLine:
        pytest -sv tests/test_explicit_profile.py -k test_explicit_profile_with_duplicate_functions
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)

        code = ub.codeblock(
            '''
            from line_profiler import profile

            @profile
            def func1(a):
                return a + 1

            @profile
            def func2(a):
                return a + 1

            @profile
            def func3(a):
                return a + 1

            @profile
            def func4(a):
                return a + 1

            func1(1)
            func2(1)
            func3(1)
            func4(1)
            ''').strip()
        with ub.ChDir(temp_dpath):

            script_fpath = ub.Path('script.py')
            script_fpath.write_text(code)

            args = [sys.executable, os.fspath(script_fpath), '--line-profile']
            proc = ub.cmd(args)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        output_fpath = (temp_dpath / 'profile_output.txt')
        raw_output = output_fpath.read_text()
        print(raw_output)

        assert 'Function: func1' in raw_output
        assert 'Function: func2' in raw_output
        assert 'Function: func3' in raw_output
        assert 'Function: func4' in raw_output

        assert output_fpath.exists()
        assert (temp_dpath / 'profile_output.lprof').exists()


def test_explicit_profile_with_customized_config():
    """
    Test that explicit profiling can be configured with the appropriate
    TOML file.
    """
    with tempfile.TemporaryDirectory() as tmp:
        temp_dpath = ub.Path(tmp)

        env = os.environ.copy()
        env['PROFILE'] = '1'

        with ub.ChDir(temp_dpath):
            script_fpath = ub.Path('script.py')
            script_fpath.write_text(_demo_explicit_profile_script())
            toml = ub.Path('my_config.toml')
            toml.write_text(ub.codeblock('''
        [tool.line_profiler.setup]
        environ_flags = ['PROFILE']

        [tool.line_profiler.write]
        output_prefix = 'my_profiling_results'
        timestamped_text = false

        [tool.line_profiler.show]
        details = true
        summarize = false
            '''))

            env['LINE_PROFILER_RC'] = str(toml)
            args = [sys.executable, os.fspath(script_fpath)]
            proc = ub.cmd(args, env=env)
            print(proc.stdout)
            print(proc.stderr)
            proc.check_returncode()

        # Check the `write` config
        assert set(os.listdir(temp_dpath)) == {'script.py',
                                               'my_config.toml',
                                               'my_profiling_results.lprof',
                                               'my_profiling_results.txt'}
        # Check the `show` config
        assert '- fib' not in proc.stdout  # No summary
        assert 'Function: fib' in proc.stdout  # With details


@pytest.mark.parametrize('reset_enable_count', [True, False])
@pytest.mark.parametrize('wrap_class, wrap_module',
                         [(None, None), (False, True),
                          (True, False), (True, True)])
def test_profiler_add_methods(wrap_class, wrap_module, reset_enable_count):
    """
    Test the `wrap` argument for the
    `LineProfiler.add_class()`, `.add_module()`, and
    `.add_imported_function_or_module()` (added via
    `line_profiler.autoprofile.autoprofile.
    _extend_line_profiler_for_profiling_imports()`) methods.
    """
    script = ub.codeblock('''
        from line_profiler import LineProfiler
        from line_profiler.autoprofile.autoprofile import (
            _extend_line_profiler_for_profiling_imports as upgrade_profiler)

        import my_module_1
        from my_module_2 import Class
        from my_module_3 import func3


        profiler = LineProfiler()
        upgrade_profiler(profiler)
        # This dispatches to `.add_module()`
        profiler.add_imported_function_or_module(my_module_1{})
        # This dispatches to `.add_class()`
        profiler.add_imported_function_or_module(Class{})
        profiler.add_imported_function_or_module(func3)

        if {}:
            for _ in range(profiler.enable_count):
                profiler.disable_by_count()

        # `func1()` should only have timing info if `wrap_module`
        my_module_1.func1()
        # `method2()` should only have timing info if `wrap_class`
        Class.method2()
        # `func3()` is profiled but don't see any timing info because it
        # isn't wrapped and doesn't auto-`.enable()` before being called
        func3()
        profiler.print_stats(details=True, summarize=True)
                          '''.format(
        '' if wrap_module is None else f', wrap={wrap_module}',
        '' if wrap_class is None else f', wrap={wrap_class}',
        reset_enable_count))

    with enter_tmpdir() as curdir:
        write(curdir / 'script.py', script)
        write(curdir / 'my_module_1.py',
              '''
        def func1():
            pass  # Marker: func1
              ''')
        write(curdir / 'my_module_2.py',
              '''
        class Class:
            @classmethod
            def method2(cls):
                pass  # Marker: method2
              ''')
        write(curdir / 'my_module_3.py',
              '''
        def func3():
            pass  # Marker: func3
              ''')
        proc = ub.cmd([sys.executable, str(curdir / 'script.py')])

    # Check that the profiler has seen each of the methods
    raw_output = proc.stdout
    print(script)
    print(raw_output)
    print(proc.stderr)
    proc.check_returncode()
    assert '# Marker: func1' in raw_output
    assert '# Marker: method2' in raw_output
    assert '# Marker: func3' in raw_output

    # Check that the timing info (of the lack thereof) are correct
    for func, has_timing in [('func1', wrap_module), ('method2', wrap_class),
                             ('func3', False)]:
        line, = (line for line in raw_output.splitlines()
                 if line.endswith('Marker: ' + func))
        has_timing = has_timing or not reset_enable_count
        assert line.split()[1] == ('1' if has_timing else 'pass')


def test_profiler_add_class_recursion_guard():
    """
    Test that if we were to add a pair of classes which each of them
    has a reference to the other in its namespace, we don't end up in
    infinite recursion.
    """
    from line_profiler import LineProfiler

    class Class1:
        def method1(self):
            pass

        class ChildClass1:
            def child_method_1(self):
                pass

    class Class2:
        def method2(self):
            pass

        class ChildClass2:
            def child_method_2(self):
                pass

        OtherClass = Class1
        # A duplicate reference shouldn't affect profiling either
        YetAnotherClass = Class1

    # Add self/mutual references
    Class1.ThisClass = Class1
    Class1.OtherClass = Class2

    profile = LineProfiler()
    profile.add_class(Class1)
    assert len(profile.functions) == 4
    assert Class1.method1 in profile.functions
    assert Class2.method2 in profile.functions
    assert Class1.ChildClass1.child_method_1 in profile.functions
    assert Class2.ChildClass2.child_method_2 in profile.functions


def test_profiler_warn_unwrappable():
    """
    Test for warnings when using `LineProfiler.add_*(wrap=True)` with a
    namespace which doesn't allow attribute assignment.
    """
    from line_profiler import LineProfiler

    class ProblamticMeta(type):
        def __init__(cls, *args, **kwargs):
            super(ProblamticMeta, cls).__init__(*args, **kwargs)
            cls._initialized = True

        def __setattr__(cls, attr, value):
            if not getattr(cls, '_initialized', None):
                return super(ProblamticMeta, cls).__setattr__(attr, value)
            raise AttributeError(
                f'cannot set attribute on {type(cls)} instance')

    class ProblematicClass(metaclass=ProblamticMeta):
        def method(self):
            pass

    profile = LineProfiler()
    vanilla_method = ProblematicClass.method

    with pytest.warns(match=r"cannot wrap 1 attribute\(s\) of "
                      r"<class '.*\.ProblematicClass'> \(`\{attr: value\}`\): "
                      r"\{'method': <function .*\.method at 0x.*>\}"):
        # The method is added to the profiler, but we can't assign its
        # wrapper back into the class namespace
        assert profile.add_class(ProblematicClass, wrap=True) == 1

    assert ProblematicClass.method is vanilla_method


@pytest.mark.parametrize(
    ('scoping_policy', 'add_module_targets', 'add_class_targets'),
    [('exact', {}, {'class3_method'}),
     ('children',
      {'class2_method', 'child_class2_method'},
      {'class3_method', 'child_class3_method'}),
     ('descendants',
      {'class2_method', 'child_class2_method',
       'class3_method', 'child_class3_method'},
      {'class3_method', 'child_class3_method'}),
     ('siblings',
      {'class1_method', 'child_class1_method',
       'class2_method', 'child_class2_method',
       'class3_method', 'child_class3_method', 'other_class3_method'},
      {'class3_method', 'child_class3_method', 'other_class3_method'}),
     ('none',
      {'class1_method', 'child_class1_method',
       'class2_method', 'child_class2_method',
       'class3_method', 'child_class3_method', 'other_class3_method'},
      {'child_class1_method',
       'class3_method', 'child_class3_method', 'other_class3_method'})])
def test_profiler_class_scope_matching(monkeypatch,
                                       scoping_policy,
                                       add_module_targets,
                                       add_class_targets):
    """
    Test for the class-scope-matching strategies of the
    `LineProfiler.add_*()` methods.
    """
    with ExitStack() as stack:
        stack.enter_context(restore_sys_modules())
        curdir = stack.enter_context(enter_tmpdir())

        pkg_dir = curdir / 'packages' / 'my_pkg'
        write(pkg_dir / '__init__.py')
        write(pkg_dir / 'submod1.py',
              """
        class Class1:
            def class1_method(self):
                pass

            class ChildClass1:
                def child_class1_method(self):
                    pass
              """)
        write(pkg_dir / 'subpkg2' / '__init__.py',
              """
        from ..submod1 import Class1  # Import from a sibling
        from .submod3 import Class3  # Import descendant from a child


        class Class2:
            def class2_method(self):
                pass

            class ChildClass2:
                def child_class2_method(self):
                    pass

            BorrowedChildClass = Class1.ChildClass1  # Non-sibling class
              """)
        write(pkg_dir / 'subpkg2' / 'submod3.py',
              """
        from ..submod1 import Class1


        class Class3:
            def class3_method(self):
                pass

            class OtherChildClass3:
                def child_class3_method(self):
                    pass

            # Unrelated class
            BorrowedChildClass1 = Class1.ChildClass1

        class OtherClass3:
            def other_class3_method(self):
                pass

        # Sibling class
        Class3.BorrowedChildClass3 = OtherClass3
              """)
        monkeypatch.syspath_prepend(pkg_dir.parent)

        from my_pkg import subpkg2
        from line_profiler import LineProfiler

        policies = {'func': 'none', 'class': scoping_policy,
                    'module': 'exact'}  # Don't descend into submodules
        # Add a module
        profile = LineProfiler()
        profile.add_module(subpkg2, scoping_policy=policies)
        assert len(profile.functions) == len(add_module_targets)
        added = {func.__name__ for func in profile.functions}
        assert added == set(add_module_targets)
        # Add a class
        profile = LineProfiler()
        profile.add_class(subpkg2.Class3, scoping_policy=policies)
        assert len(profile.functions) == len(add_class_targets)
        added = {func.__name__ for func in profile.functions}
        assert added == set(add_class_targets)


@pytest.mark.parametrize(
    ('scoping_policy', 'add_module_targets', 'add_subpackage_targets'),
    [('exact', {'func4'}, {'class_method'}),
     ('children', {'func4'}, {'class_method', 'func2'}),
     ('descendants', {'func4'}, {'class_method', 'func2'}),
     ('siblings', {'func4'}, {'class_method', 'func2', 'func3'}),
     ('none',
      {'func4', 'func5'},
      {'class_method', 'func2', 'func3', 'func4', 'func5'})])
def test_profiler_module_scope_matching(monkeypatch,
                                        scoping_policy,
                                        add_module_targets,
                                        add_subpackage_targets):
    """
    Test for the module-scope-matching strategies of the
    `LineProfiler.add_*()` methods.
    """
    with ExitStack() as stack:
        stack.enter_context(restore_sys_modules())
        curdir = stack.enter_context(enter_tmpdir())

        pkg_dir = curdir / 'packages' / 'my_pkg'
        write(pkg_dir / '__init__.py')
        write(pkg_dir / 'subpkg1' / '__init__.py',
              """
              import my_mod4  # Unrelated
              from .. import submod3  # Sibling
              from . import submod2  # Child


              class Class:
                  @classmethod
                  def class_method(cls):
                      pass

                  # We shouldn't descend into this no matter what
                  import my_mod5 as module
              """)
        write(pkg_dir / 'subpkg1' / 'submod2.py',
              """
              def func2():
                  pass
              """)
        write(pkg_dir / 'submod3.py',
              """
              def func3():
                  pass
              """)
        write(curdir / 'packages' / 'my_mod4.py',
              """
              import my_mod5  # Unrelated


              def func4():
                  pass
              """)
        write(curdir / 'packages' / 'my_mod5.py',
              """
              def func5():
                  pass
              """)
        monkeypatch.syspath_prepend(pkg_dir.parent)

        import my_mod4
        from my_pkg import subpkg1
        from line_profiler import LineProfiler

        policies = {'func': 'none', 'class': 'children',
                    'module': scoping_policy}
        # Add a module
        profile = LineProfiler()
        profile.add_module(my_mod4, scoping_policy=policies)
        assert len(profile.functions) == len(add_module_targets)
        added = {func.__name__ for func in profile.functions}
        assert added == set(add_module_targets)
        # Add a subpackage
        profile = LineProfiler()
        profile.add_module(subpkg1, scoping_policy=policies)
        assert len(profile.functions) == len(add_subpackage_targets)
        added = {func.__name__ for func in profile.functions}
        assert added == set(add_subpackage_targets)
        # Add a class
        profile = LineProfiler()
        profile.add_class(subpkg1.Class, scoping_policy=policies)
        assert [func.__name__ for func in profile.functions] == ['class_method']


@pytest.mark.parametrize(
    ('scoping_policy', 'add_module_targets', 'add_class_targets'),
    [('exact', {'func1'}, {'method'}),
     ('children', {'func1'}, {'method'}),
     ('descendants', {'func1', 'func2'}, {'method', 'child_class_method'}),
     ('siblings',
      {'func1', 'func2', 'func3'},
      {'method', 'child_class_method', 'func1'}),
     ('none',
      {'func1', 'func2', 'func3', 'func4'},
      {'method', 'child_class_method', 'func1', 'another_func4'})])
def test_profiler_func_scope_matching(monkeypatch,
                                      scoping_policy,
                                      add_module_targets,
                                      add_class_targets):
    """
    Test for the class-scope-matching strategies of the
    `LineProfiler.add_*()` methods.
    """
    with ExitStack() as stack:
        stack.enter_context(restore_sys_modules())
        curdir = stack.enter_context(enter_tmpdir())

        pkg_dir = curdir / 'packages' / 'my_pkg'
        write(pkg_dir / '__init__.py')
        write(pkg_dir / 'subpkg1' / '__init__.py',
              """
              from ..submod3 import func3  # Sibling
              from .submod2 import func2  # Descendant
              from my_mod4 import func4  # Unrelated

              def func1():
                  pass

              class Class:
                  def method(self):
                      pass

                  class ChildClass:
                      @classmethod
                      def child_class_method(cls):
                          pass

                  # Descendant
                  descdent_method = ChildClass.child_class_method

                  # Sibling
                  sibling_method = staticmethod(func1)

                  # Unrelated
                  from my_mod4 import another_func4 as imported_method
              """)
        write(pkg_dir / 'subpkg1' / 'submod2.py',
              """
              def func2():
                  pass
              """)
        write(pkg_dir / 'submod3.py',
              """
              def func3():
                  pass
              """)
        write(curdir / 'packages' / 'my_mod4.py',
              """
              def func4():
                  pass


              def another_func4(_):
                  pass
              """)
        monkeypatch.syspath_prepend(pkg_dir.parent)

        from my_pkg import subpkg1
        from line_profiler import LineProfiler

        policies = {'func': scoping_policy,
                    # No descensions
                    'class': 'exact', 'module': 'exact'}
        # Add a module
        profile = LineProfiler()
        profile.add_module(subpkg1, scoping_policy=policies)
        assert len(profile.functions) == len(add_module_targets)
        added = {func.__name__ for func in profile.functions}
        assert added == set(add_module_targets)
        # Add a class
        profile = LineProfiler()
        profile.add_module(subpkg1.Class, scoping_policy=policies)
        assert len(profile.functions) == len(add_class_targets)
        added = {func.__name__ for func in profile.functions}
        assert added == set(add_class_targets)


if __name__ == '__main__':
    ...
    test_simple_explicit_nonglobal_usage()



"""
This module defines the |lprun| and |lprun_all| IPython magic functions.

If you are using IPython, there is an implementation of an |lprun| magic
command which will let you specify functions to profile and a statement
to execute. It will also add its
:py:class:`~.LineProfiler` instance into the |builtins|, but typically,
you would not use it like that.

You can also use |lprun_all|, which profiles the whole cell you're
executing automagically, without needing to specify lines/functions
yourself. It's meant for easier use for beginners.

For IPython 0.11+, you can install it by editing the IPython configuration file
``~/.ipython/profile_default/ipython_config.py`` to add the ``'line_profiler'``
item to the extensions list::

    c.TerminalIPythonApp.extensions = [
        'line_profiler',
    ]

Or explicitly call::

    %load_ext line_profiler

To get usage help for |lprun| and |lprun_all|, use the standard IPython
help mechanism::

    In [1]: %lprun?

.. |lprun| replace:: :py:data:`%lprun <LineProfilerMagics.lprun>`
.. |lprun_all| replace:: :py:data:`%%lprun_all <LineProfilerMagics.lprun_all>`
.. |builtins| replace:: :py:mod:`__builtins__ <builtins>`
"""

import ast
import builtins
import functools
import os
import tempfile
import textwrap
import time
import types
from contextlib import ExitStack
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Union
if TYPE_CHECKING:  # pragma: no cover
    from typing import (Callable, ParamSpec,  # noqa: F401
                        Any, ClassVar, TypeVar)

    PS = ParamSpec('PS')
    PD = TypeVar('PD', bound='_PatchDict')
    DefNode = TypeVar('DefNode', ast.FunctionDef, ast.AsyncFunctionDef)

from io import StringIO

from IPython.core.getipython import get_ipython
from IPython.core.magic import Magics, magics_class, line_magic, cell_magic
from IPython.core.page import page
from IPython.utils.ipstruct import Struct
from IPython.core.error import UsageError

from line_profiler import line_profiler, LineProfiler, LineStats
from line_profiler.autoprofile.ast_tree_profiler import AstTreeProfiler


__all__ = ('LineProfilerMagics',)

_LPRUN_ALL_CODE_OBJ_NAME = '<lprof_cell>'


@dataclass
class _ParseParamResult:
    """ Class for holding parsed info relevant to the behaviors of both
    the ``%lprun`` and ``%%lprun_all`` magics.

    Attributes:
        ``.opts``
            :py:class:`IPython.utils.ipstruct.Struct` object.
        ``.arg_str``
            :py:class:`str` of unparsed argument(s).
        ``.dump_raw_dest``
            (Descriptor) :py:class:`pathlib.Path` to write the raw
            (pickled) profiling results to, or :py:data:`None` if not to
            be written.
        ``.dump_text_dest``
            (Descriptor) :py:class:`pathlib.Path` to write the
            plain-text profiling results to, or :py:data:`None` if not
            to be written.
        ``.output_unit``
            (Descriptor) Unit to normalize the output of
            :py:meth:`line_profiler.LineProfiler.print_stats` to, or
            :py:data:`None` if not specified.
        ``.strip_zero``
            (Descriptor) Whether to call
            :py:meth:`line_profiler.LineProfiler.print_stats` with
            ``stripzeros=True``.
        ``.return_profiler``
            (Descriptor) Whether the
            :py:class:`line_profiler.LineProfiler` instance is to be
            returned.
    """
    opts: Struct
    arg_str: str

    def __getattr__(self, attr):  # type: (str) -> Any
        """ Defers to :py:attr:`_ParseParamResult.opts`."""
        return getattr(self.opts, attr)

    @functools.cached_property
    def dump_raw_dest(self):  # type: () -> Path | None
        path = self.opts.D[0]
        if path:
            return Path(path)
        return None

    @functools.cached_property
    def dump_text_dest(self):  # type: () -> Path | None
        path = self.opts.T[0]
        if path:
            return Path(path)
        return None

    @functools.cached_property
    def output_unit(self):  # type: () -> float | None
        if self.opts.u is None:
            return None
        try:
            return float(self.opts.u[0])
        except Exception:
            raise TypeError("Timer unit setting must be a float.")

    @functools.cached_property
    def strip_zero(self):  # type: () -> bool
        return "z" in self.opts

    @functools.cached_property
    def return_profiler(self):  # type: () -> bool
        return "r" in self.opts


@dataclass
class _RunAndProfileResult:
    """ Class for holding the results of both the ``%lprun`` and
    ``%%lprun_all`` magics.
    """
    stats: LineStats
    parse_result: _ParseParamResult
    message: Union[str, None] = None
    time_elapsed: Union[float, None] = None
    tempfile: Union[str, 'os.PathLike[str]', None] = None

    def __post_init__(self):
        if self.tempfile is not None:
            self.tempfile = Path(self.tempfile)
        self.output  # Fetch value

    def _make_show_func_wrapper(self, show_func):
        """
        Create a replacement for
        :py:func:`line_profiler.line_profiler.show_func` to be
        monkey-patched in, so that when showing the results of the
        entire cell the lines are not truncated at the end of the first
        code block.
        """
        tmp = self.tempfile
        if tmp is None:
            return show_func
        assert isinstance(tmp, Path)

        @functools.wraps(show_func)
        def show_func_wrapper(
                filename, start_lineno, func_name, *args, **kwargs):
            call = functools.partial(show_func,
                                     filename, start_lineno, func_name,
                                     *args, **kwargs)
            show_entire_module = (start_lineno == 1
                                  and func_name == _LPRUN_ALL_CODE_OBJ_NAME
                                  and tmp is not None
                                  and tmp.samefile(filename))
            if not show_entire_module:
                return call()
            with _PatchDict.from_module(
                    line_profiler, get_code_block=get_code_block_wrapper):
                return call()

        def get_code_block_wrapper(filename, lineno):
            """ Return the entire content of :py:attr:`~.tempfile`."""
            with tmp.open(mode='r') as fobj:
                return fobj.read().splitlines(keepends=True)

        return show_func_wrapper

    @functools.cached_property
    def output(self):  # type: () -> str
        with ExitStack() as stack:
            cap = stack.enter_context(StringIO())  # Trap text output
            patch_show_func = _PatchDict.from_module(
                line_profiler,
                show_func=self._make_show_func_wrapper(line_profiler.show_func))
            stack.enter_context(patch_show_func)
            self.stats.print(cap,
                             output_unit=self.parse_result.output_unit,
                             stripzeros=self.parse_result.strip_zero)
            return cap.getvalue().rstrip()


class _PatchProfilerIntoBuiltins:
    """
    Example:
        >>> # xdoctest: +REQUIRES(module:IPython)
        >>> import builtins
        >>> from line_profiler import LineProfiler
        >>>
        >>>
        >>> prof = LineProfiler()
        >>> with _PatchProfilerIntoBuiltins(prof):
        ...     assert builtins.profile is prof
        ...
        >>> print(builtins.profile)
        Traceback (most recent call last):
          ...
        AttributeError: ...

    Note:
        This class doesn't itself need :py:mod:`IPython`, but it
        resides in a module that does. To reduce complications, we just
        skip this doctest if :py:mod:`IPython` (and hence this module)
        can't be imported.
    """
    def __init__(self, prof=None):
        # type: (LineProfiler | None) -> None
        if prof is None:
            prof = LineProfiler()
        self.prof = prof
        self._ctx = _PatchDict.from_module(builtins, profile=self.prof)

    def __enter__(self):  # type: () -> LineProfiler
        self._ctx.__enter__()
        return self.prof

    def __exit__(self, *a, **k):
        return self._ctx.__exit__(*a, **k)


class _PatchDict:
    def __init__(self, namespace, /, **kwargs):
        # type: (dict[str, Any], Any) -> None
        self.namespace = namespace
        self.replacements = kwargs
        self._stack = []  # type: list[dict[str, Any]]
        self._absent = object()

    def __enter__(self):  # type: (PD) -> PD
        self._push()
        return self

    def __exit__(self, *_, **__):
        self._pop()

    def _push(self):
        entry = {}
        namespace = self.namespace
        absent = self._absent
        for key, value in self.replacements.items():
            entry[key] = namespace.pop(key, absent)
            namespace[key] = value
        self._stack.append(entry)

    def _pop(self):
        namespace = self.namespace
        absent = self._absent
        for key, value in self._stack.pop().items():
            if value is absent:
                namespace.pop(key, None)
            else:
                namespace[key] = value

    @classmethod
    def from_module(cls, module, /, **kwargs):
        # type: (type[PD], types.ModuleType, Any) -> PD
        return cls(vars(module), **kwargs)


@magics_class
class LineProfilerMagics(Magics):
    def _parse_parameters(self, parameter_s, getopt_spec, opts_def):
        # type: (str, str, Struct) -> _ParseParamResult
        # FIXME: There is a chance that this handling will need to be
        # updated to handle single-quoted characters better (#382)
        parameter_s = parameter_s.replace('"', r"\"").replace("'", r"\"")

        opts, arg_str = self.parse_options(
            parameter_s, getopt_spec, list_all=True)
        opts.merge(opts_def)
        return _ParseParamResult(opts, arg_str)

    @staticmethod
    def _run_and_profile(prof,  # type: LineProfiler
                         parse_result,  # type: _ParseParamResult
                         tempfile,  # type: str | None
                         method,  # type: Callable[PS, Any]
                         *args,  # type: PS.args
                         **kwargs,  # type: PS.kwargs
                         ):  # type: (...) -> _RunAndProfileResult
        # Use the time module because it's easier than parsing the
        # output from `show_text()`.
        # `perf_counter()` is a monotonically increasing alternative to
        # `time()` that's intended for simple benchmarking.
        start_time = time.perf_counter()
        try:
            method(*args, **kwargs)
            message = None
        except (SystemExit, KeyboardInterrupt) as e:
            message = (f"{type(e).__name__} exception caught in "
                       "code being profiled.")

        # Capture and save total runtime
        total_time = time.perf_counter() - start_time
        return _RunAndProfileResult(
            prof.get_stats(), parse_result,
            message=message, time_elapsed=total_time, tempfile=tempfile)

    @classmethod
    def _lprun_all_get_rewritten_profiled_code(cls, tmpfile):
        # type: (str) -> types.CodeType
        """ Transform and compile the AST of the profiled code. This is
        similar to :py:meth:`.LineProfiler.runctx`,
        """
        at = AstTreeProfiler(tmpfile, [tmpfile], profile_imports=False)
        tree = at.profile()

        return compile(tree, tmpfile, "exec")

    @classmethod
    def _lprun_get_top_level_profiled_code(cls, tmpfile):
        # type: (str) -> types.CodeType
        """ Compile the profiled code."""
        with open(tmpfile, mode='r') as fobj:
            return compile(fobj.read(), tmpfile, "exec")

    @staticmethod
    def _handle_end(prof, run_result):
        # type: (LineProfiler, _RunAndProfileResult) -> LineProfiler | None
        page(run_result.output)

        dump_file = run_result.parse_result.dump_raw_dest
        if dump_file is not None:
            prof.dump_stats(dump_file)
            print(f"\n*** Profile stats pickled to file {str(dump_file)!r}.")

        text_file = run_result.parse_result.dump_text_dest
        if text_file is not None:
            with text_file.open("w", encoding="utf-8") as pfile:
                print(run_result.output, file=pfile)
            print("\n*** Profile printout saved to text file "
                  f"{str(text_file)!r}.")

        if run_result.message:
            print("\n*** " + run_result.message)

        return prof if run_result.parse_result.return_profiler else None

    @line_magic
    def lprun(self, parameter_s=""):
        """Execute a statement under the line-by-line profiler from the
        :py:mod:`line_profiler` module.

        Usage::

            %lprun [<options>] <statement>

        The given statement (which doesn't require quote marks) is run
        via the :py:class:`~.LineProfiler`. Profiling is enabled for
        the functions specified by the ``-f`` options. The statistics
        will be shown side-by-side with the code through the pager once
        the statement has completed.

        Options:

        ``-f <function>``: :py:class:`~.LineProfiler` only profiles
        functions and methods it is told to profile. This option tells
        the profiler about these functions. Multiple ``-f`` options may
        be used. The argument may be any expression that gives
        a Python function or method object. However, one must be
        careful to avoid spaces that may confuse the option parser.

        ``-m <module>``: Get all the functions/methods in a module

        One or more ``-f`` or ``-m`` options are required to get any
        useful results.

        ``-D <filename>``: dump the raw statistics out to a pickle file
        on disk. The usual extension for this is ``.lprof``. These
        statistics may be viewed later by running
        ``python -m line_profiler``.

        ``-T <filename>``: dump the text-formatted statistics with the
        code side-by-side out to a text file.

        ``-r``: return the :py:class:`~.LineProfiler` object after it
        has completed profiling.

        ``-s``: strip out all entries from the print-out that have
        zeros. This is an old, soon-to-be-deprecated alias for ``-z``.

        ``-z``: strip out all entries from the print-out that have
        zeros.

        ``-u``: specify time unit for the print-out in seconds.
        """
        opts_def = Struct(D=[""], T=[""], f=[], m=[], u=None)
        parsed = self._parse_parameters(parameter_s, "rszf:m:D:T:u:", opts_def)
        if "s" in parsed.opts:  # Handle alias
            parsed.opts["z"] = True

        assert self.shell is not None
        global_ns = self.shell.user_global_ns
        local_ns = self.shell.user_ns

        # Get the requested functions.
        funcs = []
        for name in parsed.f:
            try:
                funcs.append(eval(name, global_ns, local_ns))
            except Exception as e:
                raise UsageError(
                    f"Could not find function {name}.\n{e.__class__.__name__}: {e}"
                )

        profile = LineProfiler(*funcs)

        # Get the modules, too
        for modname in parsed.m:
            try:
                mod = __import__(modname, fromlist=[""])
                profile.add_module(mod)
            except Exception as e:
                raise UsageError(
                    f"Could not find module {modname}.\n{e.__class__.__name__}: {e}"
                )

        with _PatchProfilerIntoBuiltins(profile):
            run = self._run_and_profile(
                profile, parsed, None, profile.runctx, parsed.arg_str,
                globals=global_ns, locals=local_ns)

        return self._handle_end(profile, run)

    @cell_magic
    def lprun_all(self, parameter_s="", cell=""):
        """Execute the whole notebook cell under the line-by-line
        profiler from the :py:mod:`line_profiler` module.

        Usage::

            %%lprun_all [<options>]

        By default, without the ``-p`` option, it includes nested
        functions in the profiler. The statistics will be shown
        side-by-side with the code through the pager once the statement
        has completed.

        Options:

        ``-D <filename>``: dump the raw statistics out to a pickle file
        on disk. The usual extension for this is ``.lprof``. These
        statistics may be viewed later by running
        ``python -m line_profiler``.

        ``-T <filename>``: dump the text-formatted statistics with the
        code side-by-side out to a text file.

        ``-r``: return the :py:class:`~.LineProfiler` object after it
        has completed profiling.

        ``-z``: strip out all entries from the print-out that have
        zeros. This is included for consistency with the CLI.

        ``-u``: specify time unit for the print-out in seconds.

        ``-t``: store the total time taken (in seconds) to a variable
        called ``_total_time_taken`` in your notebook. This can be
        useful if you want to plot the total time taken for different
        versions of a code cell without needing to manually look at and
        type down the time taken. This can be accomplished with ``-r``,
        but that would require a decent bit of boilerplate code and some
        knowledge of the timings data structure, so this is added to be
        beginner-friendly.

        ``-p``: Profile only top-level code (ignore nested functions).
        Using this can bypass any issues with :py:mod:`ast`
        transformations.
        """
        opts_def = Struct(D=[""], T=[""], u=None)
        parsed = self._parse_parameters(parameter_s, "rzptD:T:u:", opts_def)

        ip = get_ipython()
        if not cell.strip():  # Edge case
            cell = "..."

        # Write the cell to a temporary file so `show_text()` inside
        # `print_stats()` can open it.
        with tempfile.NamedTemporaryFile(
            suffix=".py", delete=False, mode="w", encoding="utf-8"
        ) as tf:
            tf.write(textwrap.dedent(cell).strip('\n'))

        try:
            if "p" not in parsed.opts:  # This is the default case.
                get_code = self._lprun_all_get_rewritten_profiled_code
            else:
                get_code = self._lprun_get_top_level_profiled_code
            # Inject a fresh LineProfiler into @profile.
            with _PatchProfilerIntoBuiltins() as prof:
                code = get_code(tf.name).replace(
                    co_name=_LPRUN_ALL_CODE_OBJ_NAME)
                try:
                    code = code.replace(
                        co_qualname=_LPRUN_ALL_CODE_OBJ_NAME)
                except TypeError:  # Python < 3.11
                    pass
                # "Register" the profiled code object with the profiler
                # Notes:
                # - This uses a dummy "function" object in a hacky way,
                #   but it's OK since `add_function()` ultimately only
                #   looks at the object's `.__code__` or
                #   `.__func__.__code__`.
                # - `prof.add_function()` might have replaced the code
                #   object, so retrieve it back from the dummy function
                mock_func = types.SimpleNamespace(__code__=code)
                prof.add_function(mock_func)  # type: ignore[arg-type]
                code = mock_func.__code__
                # Notes:
                # - We don't define `ip.user_global_ns` and `ip.user_ns`
                #   at the beginning like in lprun because the ns
                #   changes after the previous compile call.
                # - The method `._run_and_profile()` fetches the
                #   `LineProfiler.print_stats()` output before the
                #   `os.unlink()` below happens, allowing for transient
                #   items to be profiled.
                with prof:
                    run = self._run_and_profile(
                        prof, parsed, tf.name, exec, code,
                        # `globals` and `locals`
                        ip.user_global_ns, ip.user_ns)
        finally:
            os.unlink(tf.name)
        if "t" in parsed.opts:
            # I know it seems redundant to include this because users
            # could just use -r to get the info, but see the docstring
            # for why -t is included anyway.
            ip.user_ns["_total_time_taken"] = run.time_elapsed

        return self._handle_end(prof, run)



import inspect


def add_imported_function_or_module(self, item, *,
                                    scoping_policy=None, wrap=False):
    """
    Method to add an object to
    :py:class:`~.line_profiler.LineProfiler` to be profiled.

    This method is used to extend an instance of
    :py:class:`~.line_profiler.LineProfiler` so it can identify whether
    an object is a callable (wrapper), a class, or a module, and handle
    its profiling accordingly.

    Args:
        item (Union[Callable, Type, ModuleType]):
            Object to be profiled.
        scoping_policy (Union[ScopingPolicy, str, ScopingPolicyDict, \
None]):
            Whether (and how) to match the scope of members and decide
            on whether to add them:

            :py:class:`str` (incl. :py:class:`~.ScopingPolicy`):
                Strings are converted to :py:class:`~.ScopingPolicy`
                instances in a case-insensitive manner, and the same
                policy applies to all members.

            ``{'func': ..., 'class': ..., 'module': ...}``
                Mapping specifying individual policies to be enacted for
                the corresponding member types.

            :py:const:`None`
                The default, equivalent to
                :py:data:`~line_profiler.line_profiler\
.DEFAULT_SCOPING_POLICIES`.

            See :py:class:`line_profiler.line_profiler.ScopingPolicy`
            and :py:meth:`~.ScopingPolicy.to_policies` for details.
        wrap (bool):
            Whether to replace the wrapped members with wrappers which
            automatically enable/disable the profiler when called.

    Returns:
        1 if any function is added to the profiler, 0 otherwise.

    See also:
        :py:data:`~line_profiler.line_profiler\
.DEFAULT_SCOPING_POLICIES`,
        :py:meth:`.LineProfiler.add_callable()`,
        :py:meth:`.LineProfiler.add_module()`,
        :py:meth:`.LineProfiler.add_class()`,
        :py:class:`~.ScopingPolicy`,
        :py:meth:`ScopingPolicy.to_policies() \
<line_profiler.line_profiler.ScopingPolicy.to_policies>`
    """
    if inspect.isclass(item):
        count = self.add_class(item, scoping_policy=scoping_policy, wrap=wrap)
    elif inspect.ismodule(item):
        count = self.add_module(item, scoping_policy=scoping_policy, wrap=wrap)
    else:
        try:
            count = self.add_callable(item)
        except TypeError:
            count = 0
    if count:
        # Session-wide enabling means that we no longer have to wrap
        # individual callables to enable/disable the profiler when
        # they're called
        self.enable_by_count()
    return 1 if count else 0



import os
import re
import shlex
from contextlib import ExitStack
from functools import partial
from pathlib import Path
from sys import stderr
from tempfile import TemporaryDirectory

import pytest
import ubelt as ub

from line_profiler import load_stats


class tempdir:
    """
    Example:
        >>> with tempdir() as td:
        ...     assert td.is_dir()
        ...     assert td.samefile('.')
        ...
        >>> assert not td.is_dir()
    """
    def __init__(self, *args, **kwargs):
        self._get_tmpdir = partial(TemporaryDirectory, *args, **kwargs)
        self._stacks = []

    def __enter__(self):  # type: () -> Path
        stack = ExitStack()
        stack.__enter__()
        tmpdir = Path(stack.enter_context(self._get_tmpdir()))
        stack.enter_context(ub.ChDir(tmpdir))
        self._stacks.append(stack)
        return tmpdir

    def __exit__(self, *_, **__):
        self._stacks.pop().close()


class _TestIPython:
    @staticmethod
    def _get_ipython_instance():
        try:
            from IPython.testing.globalipapp import get_ipython
        except ImportError:
            pytest.skip(reason='no `IPython`')
        return get_ipython()

    @staticmethod
    def _emit(request, /, *args, **kwargs):
        config = request.getfixturevalue('pytestconfig')
        # Only emit the messages when we're not capturing
        if config.getoption('capture') in (False, 'no'):
            print(*args, **kwargs)


class TestLPRun(_TestIPython):
    """
    CommandLine:
        pytest -k "TestLPRun and not TestLPRunAll" -s -v
    """
    @pytest.mark.parametrize('modules', [None, 'calendar'])
    def test_lprun_profiling_targets(self, request, modules):
        """ Test ``%lprun`` with the ``-m`` flag.
        """
        mods = shlex.split(modules or '')
        if mods:
            chunks = []
            for mod in mods:
                chunks.extend(['-m', mod])
            more_flags = shlex.join(chunks)
        else:
            more_flags = None
        lprof = self._test_lprun(request, more_flags)

        # Check the profiling of functions
        # - from the `-f` flag
        assert any(getattr(func, '__name__', None) == 'func'
                   for func in lprof.functions)
        # - from the `-m` flag
        for mod in mods:
            assert any(
                (getattr(func, '__module__', '') == mod
                 or getattr(func, '__module__', '').startswith(mod + '.'))
                for func in lprof.functions)

    @pytest.mark.parametrize(
        ('output', 'text'),
        [(None, None), ('myprof.txt', True), ('myprof.lprof', False)])
    def test_lprun_file_io(self, request, output, text):
        """ Test ``%lprun`` with the ``-D`` and ``-T`` flags.
        """
        with tempdir() as tmpdir:
            if output:
                more_flags = shlex.join(['-' + ('T' if text else 'D'), output])
            else:
                more_flags = None
            self._test_lprun(request, more_flags)

            # Check the output files (`-D` or `-T`)
            if output:
                out_path = tmpdir / output
                assert out_path.exists()
                assert out_path.stat().st_size
                if not text:  # Test roundtripping
                    load_stats(out_path)
            else:
                assert not os.listdir(tmpdir)

    @pytest.mark.parametrize('bad', [True, False])
    def test_lprun_timer_unit(self, request, bad):
        """ Test ``%lprun`` with the ``-u`` flag.
        """
        capsys = request.getfixturevalue('capsys')
        if bad:  # Test invalid value
            with pytest.raises(TypeError):
                self._test_lprun(request, '-u not_a_number')
            return
        else:
            unit = 1E-3
            self._test_lprun(request, f'-u {unit}')

        out = capsys.readouterr().out
        # Check the timer (`-u`)
        pattern = re.compile(r'Timer unit:\s*([^\s]+)\s*s')
        match, = (
            m for m in (pattern.match(line) for line in out.splitlines())
            if m)
        assert pytest.approx(float(match.group(1))) == unit

    @pytest.mark.parametrize('skip_zero', [None, '-s', '-z'])
    def test_lprun_skip_zero(self, request, skip_zero):
        """ Test ``%lprun`` with the ``-s`` and ``-z`` flags.
        """
        capsys = request.getfixturevalue('capsys')
        # Throw in an unrelated module, whose timings are always zero
        more_flags = '-m calendar'
        if skip_zero:
            more_flags = f'{more_flags} {skip_zero}'
        self._test_lprun(request, more_flags)

        # Check whether zero entries are skipped
        out = capsys.readouterr().out

        match = re.search(
            r'^File:\s*.*{}calendar\.py'.format(re.escape(os.sep)),
            out,
            flags=re.MULTILINE)
        assert bool(match) == (not skip_zero)

    @pytest.mark.parametrize(('xc', 'raised'),
                             [(SystemExit(0), False),
                              (ValueError('foo'), True)])
    def test_lprun_exception_handling(self, capsys, xc, raised):
        """ Test ``%lprun``-ing a function which raises exceptions.
        """
        ip = self._get_ipython_instance()
        ip.run_line_magic('load_ext', 'line_profiler')
        xc_repr = '{}({})'.format(
            type(xc).__name__, ', '.join(repr(a) for a in xc.args))
        ip.run_cell(
            raw_cell=f'func = lambda: (_ for _ in ()).throw({xc_repr})')

        if raised:  # Normal excepts should be bubbled up
            with pytest.raises(type(xc)):
                ip.run_line_magic('lprun', '-f func func()')
            return

        # Special expressions are captured and relegated to a warning
        # message
        ip.run_line_magic('lprun', '-f func func()')
        out = capsys.readouterr().out
        assert f'*** {type(xc).__name__} exception caught' in out

    def _test_lprun(self, request, more_flags):
        ip = self._get_ipython_instance()
        ip.run_line_magic('load_ext', 'line_profiler')
        ip.run_cell(raw_cell='def func():\n    return 2**20')
        command = '-r -f func func()'
        if more_flags:
            command = f'{more_flags} {command}'
        lprof = ip.run_line_magic('lprun', command)

        # Check the recorded timings
        filtered_timings = {
            (filename, lineno, funcname): entries
            for (filename, lineno, funcname), entries
            in lprof.get_stats().timings.items()
            if filename.startswith('<ipython')
            if filename.endswith('>')}
        assert len(filtered_timings) == 1  # 1 function

        func_data, lines_data = next(iter(filtered_timings.items()))
        self._emit(request, f'func_data={func_data}')
        self._emit(request, f'lines_data={lines_data}', file=stderr)
        assert func_data[1] == 1  # lineno of the function
        assert func_data[2] == "func"  # function name
        assert len(lines_data) == 1  # 1 line of code
        assert lines_data[0][0] == 2  # lineno
        assert lines_data[0][1] == 1  # hits

        return lprof


class TestLPRunAll(_TestIPython):
    """
    CommandLine:
        pytest -k TestLPRunAll -s -v
    """
    def test_lprun_all_autoprofile(self):
        """ Test ``%%lprun_all`` without the ``-p`` flag.
        """
        ip = self._get_ipython_instance()
        ip.run_line_magic('load_ext', 'line_profiler')
        lprof = ip.run_cell_magic(
            'lprun_all', line='-r', cell=self.lprun_all_cell_body)
        timings = lprof.get_stats().timings

        # 2 scopes: the module scope and an inner scope (Test.test)
        assert len(timings) == 2

        timings_iter = iter(timings.items())
        func_1_data, lines_1_data = next(timings_iter)
        func_2_data, lines_2_data = next(timings_iter)
        print(f'func_1_data={func_1_data}')
        print(f'lines_1_data={lines_1_data}')
        assert func_1_data[1] == 1  # lineno of the module
        assert len(lines_1_data) == 2  # only 2 lines were executed in this outer scope
        assert lines_1_data[0][0] == 1  # lineno
        assert lines_1_data[0][1] == 1  # hits

        print(f'func_2_data={func_2_data}')
        print(f'lines_2_data={lines_2_data}')
        assert func_2_data[1] == 2  # lineno of the inner function
        assert len(lines_2_data) == 5  # only 5 lines were executed in this inner scope
        assert lines_2_data[1][0] == 4  # lineno
        assert lines_2_data[1][1] == self.loops - 1  # hits

        # Check that the code is executed in the right scope and with
        # the expected side effects
        assert isinstance(ip.user_ns.get("Test"), type)
        assert ip.user_ns['z'] is None

    def test_lprun_all_autoprofile_toplevel(self):
        """ Test ``%%lprun_all`` with the ``-p`` flag.
        """
        ip = self._get_ipython_instance()
        ip.run_line_magic('load_ext', 'line_profiler')
        lprof = ip.run_cell_magic(
            'lprun_all', line='-r -p', cell=self.lprun_all_cell_body)
        timings = lprof.get_stats().timings

        # 1 scope: the module scope
        assert len(timings) == 1

        timings_iter = iter(timings.items())
        func_data, lines_data = next(timings_iter)
        print(f'func_data={func_data}')
        print(f'lines_data={lines_data}')
        assert func_data[1] == 1  # lineno of the module
        assert len(lines_data) == 2  # only 2 lines were executed in this outer scope
        assert lines_data[0][0] == 1  # lineno
        assert lines_data[0][1] == 1  # hits

        # Check that the code is executed in the right scope and with
        # the expected side effects
        assert isinstance(ip.user_ns.get("Test"), type)
        assert ip.user_ns['z'] is None

    def test_lprun_all_timetaken(self):
        """ Test ``%%lprun_all`` with the ``-t`` flag.
        """
        ip = self._get_ipython_instance()
        ip.run_line_magic('load_ext', 'line_profiler')
        result = ip.run_cell_magic(
            'lprun_all', line='-t', cell=self.lprun_all_cell_body)
        assert result is None  # No `-r` flag -> profiler not returned

        # Check that the code is executed in the right scope and with
        # the expected side effects
        assert isinstance(ip.user_ns.get("Test"), type)
        assert ip.user_ns['z'] is None
        # Check that the elapsed time is written to the right scope
        assert ip.user_ns.get("_total_time_taken", None) is not None

    # This example has 2 scopes
    # - The top level (module) scope, and
    # - The inner `Test.test()` (method) scope
    # when the `-p` flag is passed, the inner level shouldn't be
    # profiled
    loops = 20000
    lprun_all_cell_body = f"""
    class Test:
        def test(self):
            loops = {loops}
            for x in range(loops):
                y = x
                if x == (loops - 2):
                    break
    z = Test().test()
    """



# cython: language_level=3
# cython: legacy_implicit_noexcept=True
# used in _line_profiler.pyx
from libcpp.unordered_map cimport unordered_map
from cython.operator cimport dereference as deref

# long long int is at least 64 bytes assuming c99
ctypedef long long int int64

cdef extern from "Python_wrapper.h":
    ctypedef long long PY_LONG_LONG

cdef struct LastTime:
    int f_lineno
    PY_LONG_LONG time

cdef struct LineTime:
    long long code
    int lineno
    PY_LONG_LONG total_time
    long nhits

# Types used for mappings from code hash to last/line times.
ctypedef unordered_map[int64, LastTime] LastTimeMap
ctypedef unordered_map[int64, LineTime] LineTimeMap

cdef inline void last_erase_if_present(LastTimeMap* m, int64 key) noexcept:
    cdef LastTimeMap.iterator it = deref(m).find(key)
    if it != deref(m).end():
        deref(m).erase(it)

cdef inline LineTime* line_ensure_entry(LineTimeMap* m, int lineno, long long code_hash) noexcept:
    if not deref(m).count(lineno):
        deref(m)[lineno] = LineTime(code_hash, lineno, 0, 0)
    return &(deref(m)[lineno])



"""
Line Profiler
=============

The line_profiler module for doing line-by-line profiling of functions

+---------------+--------------------------------------------+
| Github        | https://github.com/pyutils/line_profiler   |
+---------------+--------------------------------------------+
| Pypi          | https://pypi.org/project/line_profiler     |
+---------------+--------------------------------------------+
| ReadTheDocs   | https://kernprof.readthedocs.io/en/latest/ |
+---------------+--------------------------------------------+


Installation
============

Releases of :py:mod:`line_profiler` and :py:mod:`kernprof` can be installed
using pip

.. code:: bash

    pip install line_profiler


The package also provides extras for optional dependencies, which can be
installed via:


.. code:: bash

    pip install line_profiler[all]


.. _line-profiler-basic-usage:

Line Profiler Basic Usage
=========================

To demonstrate line profiling, we first need to generate a Python script to
profile. Write the following code to a file called :file:`demo_primes.py`:

.. code:: python

    from line_profiler import profile


    @profile
    def is_prime(n):
        '''
        Check if the number "n" is prime, with n > 1.

        Returns a boolean, True if n is prime.
        '''
        max_val = n ** 0.5
        stop = int(max_val + 1)
        for i in range(2, stop):
            if n % i == 0:
                return False
        return True


    @profile
    def find_primes(size):
        primes = []
        for n in range(size):
            flag = is_prime(n)
            if flag:
                primes.append(n)
        return primes


    @profile
    def main():
        print('start calculating')
        primes = find_primes(100000)
        print(f'done calculating. Found {len(primes)} primes.')


    if __name__ == '__main__':
        main()


In this script we explicitly import the
:py:deco:`profile <line_profiler.explicit_profiler.GlobalProfiler>` function
from :py:mod:`line_profiler`, and then we decorate function of interest with
:py:deco:`profile`.

By default nothing is profiled when running the script.

.. code:: bash

    python demo_primes.py


The output will be

.. code::

    start calculating
    done calculating. Found 9594 primes.


The quickest way to enable profiling is to set the environment variable
:envvar:`LINE_PROFILE=1 <LINE_PROFILE>` and running your script as normal.

.. .. todo: add a link that points to docs showing all the different ways to
.. .. to enable profiling.

.. code:: bash

    LINE_PROFILE=1 python demo_primes.py

This will output 3 files: :file:`profile_output.txt`,
:file:`profile_output_<timestamp>.txt`, and :file:`profile_output.lprof`; and
:file:`stdout` will look something like:


.. code::

    start calculating
    done calculating. Found 9594 primes.
    Timer unit: 1e-09 s

      0.65 seconds - demo_primes.py:4 - is_prime
      1.47 seconds - demo_primes.py:19 - find_primes
      1.51 seconds - demo_primes.py:29 - main
    Wrote profile results to profile_output.txt
    Wrote profile results to profile_output_2023-08-12T193302.txt
    Wrote profile results to profile_output.lprof
    To view details run:
    python -m line_profiler -rtmz profile_output.lprof


The details contained in the output txt files or by running the script provided
in the output will show detailed line-by-line timing information for each
decorated function.


.. code::

    Timer unit: 1e-06 s

    Total time: 0.731624 s
    File: ./demo_primes.py
    Function: is_prime at line 4

    Line #      Hits         Time  Per Hit   % Time  Line Contents
    ==============================================================
         4                                           @profile
         5                                           def is_prime(n):
         6                                               '''
         7                                               Check if the number "n" is prime, with n > 1.
         8
         9                                               Returns a boolean, True if n is prime.
        10                                               '''
        11    100000      14178.0      0.1      1.9      max_val = n ** 0.5
        12    100000      22830.7      0.2      3.1      stop = int(max_val + 1)
        13   2755287     313514.1      0.1     42.9      for i in range(2, stop):
        14   2745693     368716.6      0.1     50.4          if n % i == 0:
        15     90406      11462.9      0.1      1.6              return False
        16      9594        922.0      0.1      0.1      return True


    Total time: 1.56771 s
    File: ./demo_primes.py
    Function: find_primes at line 19

    Line #      Hits         Time  Per Hit   % Time  Line Contents
    ==============================================================
        19                                           @profile
        20                                           def find_primes(size):
        21         1          0.2      0.2      0.0      primes = []
        22    100001      10280.4      0.1      0.7      for n in range(size):
        23    100000    1544196.6     15.4     98.5          flag = is_prime(n)
        24    100000      11375.4      0.1      0.7          if flag:
        25      9594       1853.2      0.2      0.1              primes.append(n)
        26         1          0.1      0.1      0.0      return primes


    Total time: 1.60483 s
    File: ./demo_primes.py
    Function: main at line 29

    Line #      Hits         Time  Per Hit   % Time  Line Contents
    ==============================================================
        29                                           @profile
        30                                           def main():
        31         1         14.0     14.0      0.0      print('start calculating')
        32         1    1604795.1    2e+06    100.0      primes = find_primes(100000)
        33         1         20.6     20.6      0.0      print(f'done calculating. Found {len(primes)} primes.')


See Also:

    * autoprofiling usage in: :py:mod:`line_profiler.autoprofile`

Limitations
===========

Line profiling does have limitations, and it is important to be aware of them.
Profiling multi-threaded, multi-processing, and asynchronous code may produce
unexpected or no results. All profiling also adds some amount of overhead to
the runtime, which may influence which parts of the code become bottlenecks.

Line profiler only measures the time between the start and end of a Python
call, so for benchmarking GPU code (e.g. with torch), which have asynchronous
or delayed behavior, it will only show the time to sync blocking calls in the
main thread.

Other profilers have different limitations and different trade-offs. It's good
to be aware of the right tool for the job. Here is a short list of other
profiling tools:


* `Scalene <https://github.com/plasma-umass/scalene>`_: A CPU+GPU+memory sampling based profiler.

* `PyInstrument  <https://github.com/joerick/pyinstrument>`_: A call stack profiler.

* `Yappi <https://github.com/sumerc/yappi>`_: A tracing profiler that is multithreading, asyncio and gevent aware.

* `profile / cProfile <https://docs.python.org/3/library/profile.html>`_: The builtin profile module.

* `timeit <https://docs.python.org/3/library/timeit.html>`_: The builtin timeit module for profiling single statements.

* `timerit <https://github.com/Erotemic/timerit>`_: A multi-statements alternative to the builtin :py:mod:`timeit` module.

* `torch.profiler <https://pytorch.org/docs/stable/profiler.html>`_ tools for profiling torch code.


.. .. todo: give more details on exact limitations.

"""
# Note: there are better ways to generate primes
# https://github.com/Sylhare/nprime

__submodules__ = [
    'line_profiler',
    'ipython_extension',
]

__autogen__ = """
mkinit ./line_profiler/__init__.py --relative
mkinit ./line_profiler/__init__.py --relative -w
"""


# from .line_profiler import __version__

# NOTE: This needs to be in sync with ../kernprof.py and line_profiler.py
__version__ = '5.0.1'

from .line_profiler import (LineProfiler, LineStats,
                            load_ipython_extension, load_stats, main,
                            show_func, show_text,)


from .explicit_profiler import profile


__all__ = ['LineProfiler', 'LineStats', 'line_profiler',
           'load_ipython_extension', 'load_stats', 'main', 'show_func',
           'show_text', '__version__', 'profile']



import ast
import os

from .ast_tree_profiler import AstTreeProfiler
from .util_static import modname_to_modpath, modpath_to_modname


def get_module_from_importfrom(node, module):
    r"""Resolve the full path of a relative import.

    Args:
        node (ast.ImportFrom)
            ImportFrom node
        module (str)
            Full dotted path relative to which the import is to occur

    Return:
        modname (str)
            Full path of the module from which the names are to be
            imported

    Example:
        >>> import ast
        >>> import functools
        >>> import textwrap
        >>>
        >>>
        >>> abs_import, *rel_imports = ast.parse(textwrap.dedent('''
        ... from a import b
        ... from . import b
        ... from .. import b
        ... from .baz import b
        ... from ..baz import b
        ... '''.strip('\n'))).body
        >>>
        >>>
        >>> get_module = functools.partial(
        ...     get_module_from_importfrom, module='foo.bar.foobar')
        >>> assert get_module(abs_import) == 'a'
        >>> assert get_module(rel_imports[0]) == 'foo.bar'
        >>> assert get_module(rel_imports[1]) == 'foo'
        >>> assert get_module(rel_imports[2]) == 'foo.bar.baz'
        >>> assert get_module(rel_imports[3]) == 'foo.baz'
    """
    level = node.level
    if not level:
        return node.module
    chunks = module.split('.')[:-level]
    if node.module:
        chunks.append(node.module)
    return '.'.join(chunks)


class ImportFromTransformer(ast.NodeTransformer):
    """Turn all the relative imports into absolute imports."""
    def __init__(self, module):
        self.module = module

    def visit_ImportFrom(self, node):
        level = node.level
        if not level:
            return self.generic_visit(node)
        module = get_module_from_importfrom(node, self.module)
        new_node = ast.ImportFrom(module=module, names=node.names, level=0)
        return self.generic_visit(ast.copy_location(new_node, node))


class AstTreeModuleProfiler(AstTreeProfiler):
    """Create an abstract syntax tree of an executable module and add
    profiling to it.

    Reads the module code and generates an abstract syntax tree, then adds nodes
    and/or decorators to the AST that adds the specified functions/methods,
    classes & modules in prof_mod to the profiler to be profiled.
    """
    @classmethod
    def _get_script_ast_tree(cls, script_file):
        tree = super()._get_script_ast_tree(script_file)
        # Note: don't drop the `.__init__` or `.__main__` suffix, lest
        # the relative imports fail
        module = modpath_to_modname(script_file,
                                    hide_main=False, hide_init=False)
        return ImportFromTransformer(module).visit(tree)

    @staticmethod
    def _is_main(fname):
        return os.path.basename(fname) == '__main__.py'

    @classmethod
    def _check_profile_full_script(cls, script_file, prof_mod):
        rp = os.path.realpath
        paths_to_check = {rp(script_file)}
        if cls._is_main(script_file):
            paths_to_check.add(rp(os.path.dirname(script_file)))
        paths_to_profile = {rp(mod) for mod in prof_mod}
        for mod in prof_mod:
            as_path = modname_to_modpath(mod)
            if as_path:
                paths_to_profile.add(rp(as_path))
        return bool(paths_to_check & paths_to_profile)



from functools import cached_property, partial, partialmethod
from types import (CodeType, FunctionType, MethodType,
                   BuiltinFunctionType, BuiltinMethodType,
                   ClassMethodDescriptorType, MethodDescriptorType,
                   MethodWrapperType, WrapperDescriptorType)
from typing import (TYPE_CHECKING, overload,
                    Any, Callable, Mapping, Protocol, TypeVar)
try:
    from typing import (  # type: ignore[attr-defined]  # noqa: F401
        ParamSpec)
except ImportError:  # Python < 3.10
    from typing_extensions import ParamSpec  # noqa: F401
try:
    from typing import (  # type: ignore[attr-defined]  # noqa: F401
        Self)
except ImportError:  # Python < 3.11
    from typing_extensions import Self  # noqa: F401
try:
    from typing import (  # type: ignore[attr-defined]  # noqa: F401
        TypeIs)
except ImportError:  # Python < 3.13
    from typing_extensions import TypeIs  # noqa: F401
from ._line_profiler import label


UnparametrizedCallableLike = TypeVar('UnparametrizedCallableLike',
                                     FunctionType, property, MethodType)
T = TypeVar('T')
T_co = TypeVar('T_co', covariant=True)
PS = ParamSpec('PS')

if TYPE_CHECKING:
    class CythonCallable(Protocol[PS, T_co]):
        def __call__(self, *args: PS.args, **kwargs: PS.kwargs) -> T_co:
            ...

        @property
        def __code__(self) -> CodeType:
            ...

        @property
        def func_code(self) -> CodeType:
            ...

        @property
        def __name__(self) -> str:
            ...

        @property
        def func_name(self) -> str:
            ...

        @property
        def __qualname__(self) -> str:
            ...

        @property
        def __doc__(self) -> str | None:
            ...

        @__doc__.setter
        def __doc__(self, doc: str | None) -> None:
            ...

        @property
        def func_doc(self) -> str | None:
            ...

        @property
        def __globals__(self) -> dict[str, Any]:
            ...

        @property
        def func_globals(self) -> dict[str, Any]:
            ...

        @property
        def __dict__(self) -> dict[str, Any]:
            ...

        @__dict__.setter
        def __dict__(self, dict: dict[str, Any]) -> None:
            ...

        @property
        def func_dict(self) -> dict[str, Any]:
            ...

        @property
        def __annotations__(self) -> dict[str, Any]:
            ...

        @__annotations__.setter
        def __annotations__(self, annotations: dict[str, Any]) -> None:
            ...

        @property
        def __defaults__(self):
            ...

        @property
        def func_defaults(self):
            ...

        @property
        def __kwdefaults__(self):
            ...

        @property
        def __closure__(self):
            ...

        @property
        def func_closure(self):
            ...


else:
    CythonCallable = type(label)

CLevelCallable = TypeVar('CLevelCallable',
                         BuiltinFunctionType, BuiltinMethodType,
                         ClassMethodDescriptorType, MethodDescriptorType,
                         MethodWrapperType, WrapperDescriptorType)


def is_c_level_callable(func: Any) -> TypeIs[CLevelCallable]:
    ...


def is_cython_callable(func: Any) -> TypeIs[CythonCallable]:
    ...


def is_classmethod(f: Any) -> TypeIs[classmethod]:
    ...


def is_staticmethod(f: Any) -> TypeIs[staticmethod]:
    ...


def is_boundmethod(f: Any) -> TypeIs[MethodType]:
    ...


def is_partialmethod(f: Any) -> TypeIs[partialmethod]:
    ...


def is_partial(f: Any) -> TypeIs[partial]:
    ...


def is_property(f: Any) -> TypeIs[property]:
    ...


def is_cached_property(f: Any) -> TypeIs[cached_property]:
    ...


class ByCountProfilerMixin:
    def get_underlying_functions(self, func) -> list[FunctionType]:
        ...

    @overload
    def wrap_callable(self,  # type: ignore[overload-overlap]
                      func: CLevelCallable) -> CLevelCallable:
        ...

    @overload
    def wrap_callable(  # type: ignore[overload-overlap]
        self, func: UnparametrizedCallableLike,
    ) -> UnparametrizedCallableLike:
        ...

    @overload
    def wrap_callable(self,  # type: ignore[overload-overlap]
                      func: type[T]) -> type[T]:
        ...

    @overload
    def wrap_callable(self,  # type: ignore[overload-overlap]
                      func: partial[T]) -> partial[T]:
        ...

    @overload
    def wrap_callable(self, func: partialmethod[T]) -> partialmethod[T]:
        ...

    @overload
    def wrap_callable(self,
                      func: cached_property[T_co]) -> cached_property[T_co]:
        ...

    @overload
    def wrap_callable(self,  # type: ignore[overload-overlap]
                      func: staticmethod[PS, T_co]) -> staticmethod[PS, T_co]:
        ...

    @overload
    def wrap_callable(
        self, func: classmethod[type[T], PS, T_co],
    ) -> classmethod[type[T], PS, T_co]:
        ...

    # Fallback: just return a wrapper function around a generic callable

    @overload
    def wrap_callable(self, func: Callable) -> FunctionType:
        ...

    def wrap_classmethod(
        self, func: classmethod[type[T], PS, T_co],
    ) -> classmethod[type[T], PS, T_co]:
        ...

    def wrap_staticmethod(
            self, func: staticmethod[PS, T_co]) -> staticmethod[PS, T_co]:
        ...

    def wrap_boundmethod(self, func: MethodType) -> MethodType:
        ...

    def wrap_partialmethod(self, func: partialmethod[T]) -> partialmethod[T]:
        ...

    def wrap_partial(self, func: partial[T]) -> partial[T]:
        ...

    def wrap_property(self, func: property) -> property:
        ...

    def wrap_cached_property(
            self, func: cached_property[T_co]) -> cached_property[T_co]:
        ...

    def wrap_async_generator(self, func: FunctionType) -> FunctionType:
        ...

    def wrap_coroutine(self, func: FunctionType) -> FunctionType:
        ...

    def wrap_generator(self, func: FunctionType) -> FunctionType:
        ...

    def wrap_function(self, func: Callable) -> FunctionType:
        ...

    def wrap_class(self, func: type[T]) -> type[T]:
        ...

    def run(self, cmd: str) -> Self:
        ...

    def runctx(self,
               cmd: str,
               globals: dict[str, Any] | None,
               locals: Mapping[str, Any] | None) -> Self:
        ...

    def runcall(self, func: Callable[PS, T], /,
                *args: PS.args, **kw: PS.kwargs) -> T:
        ...

    def __enter__(self) -> Self:
        ...

    def __exit__(self, *_, **__) -> None:
        ...



"""
New in ``line_profiler`` version 4.1.0, this module defines a top-level
``profile`` decorator which will be disabled by default **unless** a script is
being run with :mod:`kernprof`, if the environment variable
:envvar:`LINE_PROFILE` is set, or if ``--line-profile`` is given on the command
line.

In the latter two cases, the :mod:`atexit` module is used to display and dump
line profiling results to disk when Python exits.

If none of the enabling conditions are met, then
:py:obj:`line_profiler.profile` is a no-op. This means you no longer have to add
and remove the implicit ``profile`` decorators required by previous version of
this library.

Basic usage is to import line_profiler and decorate your function with
line_profiler.profile.  By default this does nothing, it's a no-op decorator.
However, if you run with the environment variable ``LINE_PROFILER=1`` or if
``'--profile' in sys.argv``, then it enables profiling and at the end of your
script it will output the profile text.

Here is a minimal example that will write a script to disk and then run it
with profiling enabled or disabled by various methods:

.. code:: bash

    # Write demo python script to disk
    python -c "if 1:
        import textwrap
        text = textwrap.dedent(
            '''
            from line_profiler import profile

            @profile
            def plus(a, b):
                return a + b

            @profile
            def fib(n):
                a, b = 0, 1
                while a < n:
                    a, b = b, plus(a, b)

            @profile
            def main():
                import math
                import time
                start = time.time()

                print('start calculating')
                while time.time() - start < 1:
                    fib(10)
                    math.factorial(1000)
                print('done calculating')

            main()
            '''
        ).strip()
        with open('demo.py', 'w') as file:
            file.write(text)
    "

    echo "---"
    echo "## Base Case: Run without any profiling"
    python demo.py

    echo "---"
    echo "## Option 0: Original Usage"
    python -m kernprof -l demo.py
    python -m line_profiler -rmt demo.py.lprof

    echo "---"
    echo "## Option 1: Enable profiler with the command line"
    python demo.py --line-profile

    echo "---"
    echo "## Option 1: Enable profiler with an environment variable"
    LINE_PROFILE=1 python demo.py


The explicit :py:attr:`line_profiler.profile` decorator can also be enabled and
configured in the Python code itself by calling
:func:`line_profiler.profile.enable`. The following example demonstrates this:

.. code:: bash

    # In-code enabling
    python -c "if 1:
        import textwrap
        text = textwrap.dedent(
            '''
            from line_profiler import profile
            profile.enable(output_prefix='customized')

            @profile
            def fib(n):
                a, b = 0, 1
                while a < n:
                    a, b = b, a + b

            fib(100)
            '''
        ).strip()
        with open('demo.py', 'w') as file:
            file.write(text)
    "
    echo "## Configuration handled inside the script"
    python demo.py


Likewise there is a :func:`line_profiler.profile.disable` function that will
prevent any subsequent functions decorated with ``@profile`` from being
profiled. In the following example, profiling information will only be recorded
for ``func2`` and ``func4``.

.. code:: bash

    # In-code enabling / disable
    python -c "if 1:
        import textwrap
        text = textwrap.dedent(
            '''
            from line_profiler import profile

            @profile
            def func1():
                return list(range(100))

            profile.enable(output_prefix='custom')

            @profile
            def func2():
                return tuple(range(100))

            profile.disable()

            @profile
            def func3():
                return set(range(100))

            profile.enable()

            @profile
            def func4():
                return dict(zip(range(100), range(100)))

            print(type(func1()))
            print(type(func2()))
            print(type(func3()))
            print(type(func4()))
            '''
        ).strip()
        with open('demo.py', 'w') as file:
            file.write(text)
    "

    echo "---"
    echo "## Configuration handled inside the script"
    python demo.py

    # Running with --line-profile will also profile ``func1``
    python demo.py --line-profile

The core functionality in this module was ported from :mod:`xdev`.
"""
from __future__ import annotations
import atexit
import multiprocessing
import os
import pathlib
import sys
import typing
from typing import Any, Callable

if typing.TYPE_CHECKING:
    ConfigArg = str | pathlib.PurePath | bool | None


# This is for compatibility
from .cli_utils import boolean, get_python_executable as _python_command
from .line_profiler import LineProfiler
from .toml_config import ConfigSource

# The first process that enables profiling records its PID here. Child processes
# created via multiprocessing (spawn/forkserver) inherit this environment value,
# which helps prevent helper processes from claiming ownership and clobbering
# output. Standalone subprocess runs should always be able to reset this value.
_OWNER_PID_ENVVAR: str = 'LINE_PROFILER_OWNER_PID'


class GlobalProfiler:
    """
    Manages a profiler that will output on interpreter exit.

    The :py:obj:`line_profile.profile` decorator is an instance of this object.

    Arguments:
        config (str | PurePath | bool | None):
            Optional TOML config file from which to load the
            configurations (see Attributes);
            if not explicitly given (= :py:data:`True` or
            :py:data:`None`), it is either resolved from the
            :envvar:`!LINE_PROFILER_RC` environment variable or looked
            up among the current directory or its ancestors.  Should all
            that fail, the default config file at
            ``importlib.resources.path('line_profiler.rc', 'line_profiler.toml')`` is used;
            passing :py:data:`False` disables all lookup and falls back
            to the default configuration

    Attributes:
        setup_config (Dict[str, List[str]]):
            Determines how the implicit setup behaves by defining which
            environment variables / command line flags to look for.
            Defaults to the ``[tool.line_profiler.setup]`` table of the
            loaded config file.

        output_prefix (str):
            The prefix of any output files written. Should include
            a part of a filename. Defaults to the ``output_prefix``
            value in the ``[tool.line_profiler.write]`` table of the
            loaded config file.

        write_config (Dict[str, bool]):
            Which outputs are enabled;
            options are lprof, text, timestamped_text, and stdout.
            Defaults to the rest of the ``[tool.line_profiler.write]``
            table of the loaded config file.

        show_config (Dict[str, bool]):
            Display configuration options;
            some outputs force certain options (e.g. text always has
            details and is never rich).
            Defaults to the rest of the ``[tool.line_profiler.show]``
            table of the loaded config file, excluding the
            ``[tool.line_profiler.show.column_widths]`` subtable.

        enabled (bool | None):
            True if the profiler is enabled (i.e. if it will wrap a function
            that it decorates with a real profiler). If None, then the value
            defaults based on the ``setup_config``, :py:obj:`os.environ`, and
            :py:obj:`sys.argv`.

    Example:
        >>> from line_profiler.explicit_profiler import *  # NOQA
        >>> self = GlobalProfiler()
        >>> # Setting the _profile attribute prevents atexit from running.
        >>> self._profile = LineProfiler()
        >>> # User can personalize the configuration
        >>> self.show_config['details'] = True
        >>> self.write_config['lprof'] = False
        >>> self.write_config['text'] = False
        >>> self.write_config['timestamped_text'] = False
        >>> # Demo data: a function to profile
        >>> def collatz(n):
        ...     while n != 1:
        ...         if n % 2 == 0:
        ...             n = n // 2
        ...         else:
        ...             n = 3 * n + 1
        ...     return n
        >>> # Disabled by default, implicitly checks to auto-enable on first wrap
        >>> assert self.enabled is None
        >>> wrapped = self(collatz)
        >>> assert self.enabled is False
        >>> assert wrapped is collatz
        >>> # Can explicitly enable
        >>> self.enable()
        >>> wrapped = self(collatz)
        >>> assert self.enabled is True
        >>> assert wrapped is not collatz
        >>> wrapped(100)
        >>> # Can explicitly request output
        >>> self.show()
    """

    _config: pathlib.PurePath | None
    _profile: LineProfiler | None
    _owner_pid: int | None
    enabled: bool | None

    setup_config: dict[str, list[str]]
    write_config: dict[str, Any]
    show_config: dict[str, Any]
    output_prefix: str

    def __init__(self, config: ConfigArg = None) -> None:
        # Remember which config file we loaded settings from
        config_source = ConfigSource.from_config(config)
        self._config = config_source.path

        self._profile = None
        self._owner_pid = None
        self.enabled = None
        # Configs:
        # - How to toggle the profiler
        self.setup_config = config_source.conf_dict['setup']
        # - Which outputs to write on exit
        self.write_config = config_source.conf_dict['write']
        # - Whither to write output files
        self.output_prefix = self.write_config.pop('output_prefix')
        # - How output will be displayed
        self.show_config = config_source.conf_dict['show']
        # (This is not stored here nor is accepted by any method, but is
        # re-parsed by `LineProfiler.print_stats()` etc. from the
        # supplied `config`)
        self.show_config.pop('column_widths')

    def _kernprof_overwrite(self, profile: LineProfiler) -> None:
        """
        Kernprof will call this when it runs, so we can use its profile object
        instead of our own. Note: when kernprof overwrites us we wont register
        an atexit hook. This is what we want because kernprof wants us to use
        another program to read its output file.
        """
        self._profile = profile
        self.enabled = True

    def _implicit_setup(self) -> None:
        """
        Called once the first time the user decorates a function with
        ``line_profiler.profile`` and they have not explicitly setup the global
        profiling options.
        """
        environ_flags = self.setup_config['environ_flags']
        cli_flags = self.setup_config['cli_flags']
        is_profiling = any(boolean(os.environ.get(f, ''), fallback=True)
                           for f in environ_flags)
        is_profiling |= any(f in sys.argv for f in cli_flags)
        if is_profiling:
            self.enable()
        else:
            self.disable()

    def enable(self, output_prefix: str | None = None) -> None:
        """
        Explicitly enables global profiler and controls its settings.

        Notes:
            Multiprocessing start methods like 'spawn'/'forkserver' can create
            helper/bootstrap interpreters that import this module. Those helpers
            must not claim ownership or register an atexit hook, otherwise they can
            clobber output from the real script process.
        """
        self._debug("enable:ENTER")

        if is_mp_bootstrap():
            self._debug("enable:skip-mp-bootstrap")
            self.enabled = False
            return

        if self._should_skip_due_to_owner():
            self._debug("enable:skip-due-to-owner")
            self.enabled = False
            return

        owner_pid = os.getpid()
        os.environ[_OWNER_PID_ENVVAR] = str(owner_pid)
        self._owner_pid = owner_pid
        self._debug("enable:owner-claimed", owner_pid=owner_pid)

        if self._profile is None:
            atexit.register(self.show)
            self._profile = LineProfiler()

        self.enabled = True
        if output_prefix is not None:
            self.output_prefix = output_prefix

    def _should_skip_due_to_owner(self) -> bool:
        """
        Return True if another process has already claimed ownership.

        The first process to enable profiling records its PID in an env var.
        Child interpreters can inherit that value; they must not steal ownership.
        """
        owner = os.environ.get(_OWNER_PID_ENVVAR)
        if not owner:
            self._debug("owner:no-owner-env")
            return False

        current = str(os.getpid())
        if owner == current:
            self._debug("owner:is-us", owner=owner)
            return False

        if is_mp_bootstrap():
            self._debug("owner:skip-mp-bootstrap", owner=owner, current=current)
            return True

        # Standalone run: allow this interpreter to become the owner.
        self._debug("owner:allow-standalone-reset", owner=owner, current=current)
        return False

    def _debug(self, message: str, **extra: Any) -> None:
        if not os.environ.get('LINE_PROFILER_DEBUG'):
            return
        try:
            parent = multiprocessing.parent_process()
            parent_pid = parent.pid if parent is not None else None
        except Exception:
            parent_pid = None

        pid = os.getpid()

        info: dict[str, Any] = {
            'ppid': os.getppid(),
            'process': getattr(multiprocessing.current_process(), 'name', None),
            'parent_pid': parent_pid,
            'owner_env': os.environ.get(_OWNER_PID_ENVVAR),
            'owner_pid': self._owner_pid,
            'enabled': self.enabled,
        }
        info.update(extra)
        payload = ' '.join(f'{k}={v!r}' for k, v in info.items())
        print(f'[line_profiler debug {pid=}] {message} {payload}')

    def disable(self) -> None:
        """
        Explicitly initialize and disable this global profiler.
        """
        self.enabled = False

    def __call__(self, func: Callable) -> Callable:
        """
        If the global profiler is enabled, decorate a function to start the
        profiler on function entry and stop it on function exit. Otherwise
        return the input.

        Args:
            func (Callable): the function to profile

        Returns:
            Callable: a potentially wrapped function
        """
        # from multiprocessing import current_process
        # if current_process().name != 'MainProcess':
        #     return func

        if self.enabled is None:
            # Force a setup if we haven't done it before.
            self._implicit_setup()
        if not self.enabled:
            return func
        assert self._profile is not None

        wrapped = self._profile(func)
        return wrapped

    def show(self) -> None:
        """
        Write the managed profiler stats to enabled outputs.

        If the implicit setup triggered, then this will be called by
        :py:mod:`atexit`.
        """
        self._debug('show:enter')
        owner_env = os.environ.get(_OWNER_PID_ENVVAR)
        if os.getppid() == 1 and owner_env == str(os.getpid()):
            self._debug('show:skip-orphan-owner', owner_env=owner_env)
            return
        if self._owner_pid is not None and os.getpid() != self._owner_pid:
            self._debug('show:skip-non-owner', current_pid=os.getpid())
            return
        import io

        write_stdout = self.write_config['stdout']
        write_text = self.write_config['text']
        write_timestamped_text = self.write_config['timestamped_text']
        write_lprof = self.write_config['lprof']

        assert self._profile is not None

        kwargs: dict[str, Any] = {'config': self._config, **self.show_config}
        if write_stdout:
            self._profile.print_stats(**kwargs)

        if write_text or write_timestamped_text:
            stream = io.StringIO()
            # Text output always contains details, and cannot be rich.
            text_kwargs: dict[str, Any] = {**kwargs, 'rich': False, 'details': True}
            self._profile.print_stats(stream=stream, **text_kwargs)
            raw_text = stream.getvalue()

            if write_text:
                txt_output_fpath1 = pathlib.Path(f'{self.output_prefix}.txt')
                txt_output_fpath1.write_text(raw_text, encoding='utf-8')
                print('Wrote profile results to %s' % txt_output_fpath1)

            if write_timestamped_text:
                from datetime import datetime as datetime_cls
                now = datetime_cls.now()
                timestamp = now.strftime('%Y-%m-%dT%H%M%S')
                txt_output_fpath2 = pathlib.Path(
                    f'{self.output_prefix}_{timestamp}.txt')
                txt_output_fpath2.write_text(raw_text, encoding='utf-8')
                print('Wrote profile results to %s' % txt_output_fpath2)

        if write_lprof:
            lprof_output_fpath = pathlib.Path(f'{self.output_prefix}.lprof')
            self._profile.dump_stats(lprof_output_fpath)
            print('Wrote profile results to %s' % lprof_output_fpath)
            print('To view details run:')
            py_exe = _python_command()
            print(py_exe + ' -m line_profiler -rtmz '
                  + str(lprof_output_fpath))


def is_mp_bootstrap() -> bool:
    """
    True when this interpreter invocation looks like multiprocessing
    bootstrapping/plumbing, where we must not claim ownership / write outputs.

    Example:
        >>> # xdoctest: +SKIP('can be flaky at test time')
        >>> import pytest
        >>> if is_mp_bootstrap():
        ...     pytest.skip('Cannot test mp bootstrap detection from within an mp bootstrap process')
        >>> import sys, subprocess, textwrap
        >>> code = textwrap.dedent(r'''
        ... import multiprocessing as mp
        ... from line_profiler.explicit_profiler import is_mp_bootstrap
        ...
        ... def child(q):
        ...     q.put(is_mp_bootstrap())
        ...
        ... if __name__ == "__main__":
        ...     ctx = mp.get_context("spawn")
        ...     q = ctx.Queue()
        ...     p = ctx.Process(target=child, args=(q,))
        ...     p.start()
        ...     val = q.get()
        ...     p.join()
        ...     print(val)
        ... ''')
        >>> out = subprocess.check_output([sys.executable, "-c", code], text=True).strip()
        >>> out in {"True", "False"}
        True
    """
    try:
        import multiprocessing.spawn as mp_spawn
        if getattr(mp_spawn, "_inheriting", False):
            return True
    except Exception:
        pass

    orig = getattr(sys, "orig_argv", None) or []
    if any(a.startswith("--multiprocessing") for a in orig):
        return True
    if any("multiprocessing.forkserver" in a for a in orig):
        return True
    if any("multiprocessing.spawn" in a for a in orig):
        return True

    try:
        if multiprocessing.current_process().name != "MainProcess":
            return True
    except Exception:
        pass

    return False


# Construct the global profiler.
# The first time it is called, it will be initialized. This is usually a
# NoOpProfiler unless the user requested the real one.
# NOTE: kernprof or the user may explicitly setup the global profiler.
profile = GlobalProfiler()



import contextlib
import os
import re
import shlex
import subprocess
import sys
import tempfile
import unittest

import pytest
import ubelt as ub

from kernprof import main, ContextualProfile


def f(x):
    """ A function. """
    y = x + 10
    return y


def g(x):
    """ A generator. """
    y = yield x + 10
    yield y + 20


@pytest.mark.parametrize(
    'use_kernprof_exec, args, expected_output, expect_error',
    [([False, ['-m'], '', True]),
     # `python -m kernprof`
     (False, ['-m', 'mymod'], "[__MYMOD__]", False),
     # `kernprof`
     (True, ['-m', 'mymod'], "[__MYMOD__]", False),
     (False, ['-m', 'mymod', '-p', 'bar'], "[__MYMOD__, '-p', 'bar']", False),
     # `-p bar` consumed by `kernprof`, `-p baz` are not
     (False,
      ['-p', 'bar', '-m', 'mymod', '-p', 'baz'],
      "[__MYMOD__, '-p', 'baz']",
      False),
     # Separator `--` broke off the remainder, so the requisite arg for
     # `-m` is not found and we error out
     (False, ['-p', 'bar', '-m', '--', 'mymod', '-p', 'baz'], '', True),
     # Separator `--` broke off the remainder, so `-m` is passed to the
     # script instead of being parsed as the module to execute
     (False,
      ['-p', 'bar', 'mymod.py', '--', '-m', 'mymod', '-p', 'baz'],
      "['mymod.py', '-m', 'mymod', '-p', 'baz']",
      False)])
def test_kernprof_m_parsing(
        use_kernprof_exec, args, expected_output, expect_error):
    """
    Test that `kernprof -m` behaves like `python -m` in that it requires
    an argument and cuts off everything after it, passing that along
    to the module to be executed.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_dpath = ub.Path(tmpdir)
        mod = (temp_dpath / 'mymod.py').resolve()
        mod.write_text(ub.codeblock(
            '''
            import sys


            if __name__ == '__main__':
                print(sys.argv)
            '''))
        if use_kernprof_exec:
            cmd = ['kernprof']
        else:
            cmd = [sys.executable, '-m', 'kernprof']
        proc = ub.cmd(cmd + args, cwd=temp_dpath, verbose=2)
    if expect_error:
        assert proc.returncode
        return
    else:
        proc.check_returncode()
    expected_output = re.escape(expected_output).replace(
        '__MYMOD__', "'.*{}'".format(re.escape(os.path.sep + mod.name)))
    assert re.match('^' + expected_output, proc.stdout)


@pytest.mark.skipif(sys.version_info[:2] < (3, 11),
                    reason='no `@enum.global_enum` in Python '
                    f'{".".join(str(v) for v in sys.version_info[:3])}')
@pytest.mark.parametrize(('flags', 'profiled_main'),
                         [('-lv -p mymod', True),  # w/autoprofile
                          ('-lv', True),  # w/o autoprofile
                          ('-b', False)])  # w/o line profiling
def test_kernprof_m_sys_modules(flags, profiled_main):
    """
    Test that `kernprof -m` is amenable to modules relying on the global
    `sys` state (e.g. those using `@enum.global_enum`).
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'mymod.py').write_text(ub.codeblock(
            '''
            import enum
            import os
            import sys


            @enum.global_enum
            class MyEnum(enum.Enum):
                FOO = 1
                BAR = 2


            @profile
            def main():
                x = FOO.value + BAR.value
                print(x)
                assert x == 3


            if __name__ == '__main__':
                main()
            '''))
        cmd = [sys.executable, '-m', 'kernprof',
               *shlex.split(flags), '-m', 'mymod']
        proc = ub.cmd(cmd, cwd=temp_dpath, verbose=2)
    proc.check_returncode()
    assert proc.stdout.startswith('3\n')
    assert ('Function: main' in proc.stdout) == profiled_main


@pytest.mark.parametrize('autoprof', [True, False])
@pytest.mark.parametrize('static', [True, False])
def test_kernprof_m_import_resolution(static, autoprof):
    """
    Test that `kernprof -m` resolves the module using static and dynamic
    as is appropriate (toggled by the undocumented environment variable
    :env:`LINE_PROFILER_STATIC_ANALYSIS`; note that static analysis
    doesn't handle namespace modules while dynamic resolution does.
    """
    code = ub.codeblock('''
    import enum
    import os
    import sys


    @profile
    def main():
        print('Hello world')


    if __name__ == '__main__':
        main()
    ''')
    cmd = [sys.executable, '-m', 'kernprof', '-lv']
    if autoprof:
        # Remove the explicit decorator, and use the `--prof-mod` flag
        code = '\n'.join(line for line in code.splitlines()
                         if '@profile' not in line)
        cmd += ['-p', 'my_namesapce_pkg.mysubmod']
    with tempfile.TemporaryDirectory() as tmpdir:
        temp_dpath = ub.Path(tmpdir)
        namespace_mod_path = temp_dpath / 'my_namesapce_pkg' / 'mysubmod.py'
        namespace_mod_path.parent.mkdir()
        namespace_mod_path.write_text(code)
        python_path = tmpdir
        if 'PYTHONPATH' in os.environ:
            python_path += ':' + os.environ['PYTHONPATH']
        env = {**os.environ,
               # Toggle use of static analysis
               'LINE_PROFILER_STATIC_ANALYSIS': str(bool(static)),
               # Add the tempdir to `sys.path`
               'PYTHONPATH': python_path}
        cmd += ['-m', 'my_namesapce_pkg.mysubmod']
        proc = ub.cmd(cmd, cwd=temp_dpath, verbose=2, env=env)
    if static:
        assert proc.returncode
        assert proc.stderr.startswith('Could not find module')
    else:
        proc.check_returncode()
        assert proc.stdout.startswith('Hello world\n')
        assert 'Function: main' in proc.stdout


@pytest.mark.parametrize('error', [True, False])
@pytest.mark.parametrize(
    'args',
    ['', '-pmymod'],  # Normal execution / auto-profile
)
def test_kernprof_sys_restoration(capsys, error, args):
    """
    Test that `kernprof.main()` and
    `line_profiler.autoprofile.autoprofile.run()` (resp.) properly
    restores `sys.path` (resp. `sys.modules['__main__']`) on the way
    out.

    Notes
    -----
    The test is run in-process.
    """
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        tmpdir = enter(tempfile.TemporaryDirectory())
        assert tmpdir not in sys.path
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'mymod.py').write_text(ub.codeblock(
            f'''
            import sys


            def main():
                # Mess up `sys.path`
                sys.path.append({tmpdir!r})
                # Output
                print(1)
                # Optionally raise an error
                if {error!r}:
                    raise Exception


            if __name__ == '__main__':
                main()
            '''))
        enter(ub.ChDir(tmpdir))
        if error:
            ctx = pytest.raises(BaseException)
        else:
            ctx = contextlib.nullcontext()
        old_modules = sys.modules.copy()
        try:
            old_main = sys.modules.get('__main__')
            with ctx:
                main(['-l', *shlex.split(args), '-m', 'mymod'])
            out, _ = capsys.readouterr()
            assert out.startswith('1')
            assert tmpdir not in sys.path
            assert sys.modules.get('__main__') is old_main
        finally:
            sys.modules.clear()
            sys.modules.update(old_modules)


@pytest.mark.parametrize(
    ('flags', 'expected_stdout', 'expected_stderr'),
    [('',  # Neutral verbosity level
      {'^Output to stdout': True,
       r"^Wrote .* '.*script\.py\.lprof'": True,
       r'^Inspect results with:''\n'
       r'\S*python\S* -m line_profiler .*script\.py\.lprof': True,
       r'line_profiler\.autoprofile\.autoprofile'
       r'\.run\(\n(?:.+,\n)*.*\)': False,
       r'^\[kernprof .*\]': False,
       r'^Function: main': False},
      {'^Output to stderr': True}),
     ('--view',  # Verbosity level 1 = `--view`
      {'^Output to stdout': True,
       r"^Wrote .* '.*script\.py\.lprof'": True,
       r'^Inspect results with:''\n'
       r'\S*python\S* -m line_profiler .*script\.py\.lprof': False,
       r'line_profiler\.autoprofile\.autoprofile'
       r'\.run\(\n(?:.+,\n)*.*\)': False,
       r'^\[kernprof .*\]': False,
       r'^Function: main': True},
      {'^Output to stderr': True}),
     ('-vv',  # Verbosity level 2, show diagnostics
      {'^Output to stdout': True,
       r"^\[kernprof .*\] Wrote .* '.*script\.py\.lprof'": True,
       r'Inspect results with:''\n'
       r'\S*python\S* -m line_profiler .*script\.py\.lprof': False,
       r'line_profiler\.autoprofile\.autoprofile'
       r'\.run\(\n(?:.+,\n)*.*\)': True,
       r'^Function: main': True},
      {'^Output to stderr': True}),
     # Verbosity level -1, suppress `kernprof` output
     ('--quiet',
      {'^Output to stdout': True, 'Wrote': False},
      {'^Output to stderr': True}),
     # Verbosity level -2, suppress script stdout
     # (also test verbosity arithmatics)
     ('--quiet --quiet --verbose -q', None, {'^Output to stderr': True}),
     # Verbosity level -3, suppress script stderr
     ('-qq --quiet', None,
      # This should have been `None`, but there's something weird with
      # `coverage` in CI which causes a spurious warning...
      {'^Output to stderr': False})])
def test_kernprof_verbosity(flags, expected_stdout, expected_stderr):
    """
    Test the various verbosity levels of `kernprof`.
    """
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        tmpdir = enter(tempfile.TemporaryDirectory())
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'script.py').write_text(ub.codeblock(
            '''
            import sys


            def main():
                print('Output to stdout', file=sys.stdout)
                print('Output to stderr', file=sys.stderr)


            if __name__ == '__main__':
                main()
            '''))
        enter(ub.ChDir(tmpdir))
        proc = ub.cmd(['kernprof', '-l',
                       # Add an eager pre-import target
                       '-p', 'script.py', '-p', 'zipfile', '-z',
                       *shlex.split(flags), 'script.py'])
    proc.check_returncode()
    print(proc.stdout)
    for expected_outputs, stream in [(expected_stdout, proc.stdout),
                                     (expected_stderr, proc.stderr)]:
        if expected_outputs is None:
            assert not stream
            continue
        for pattern, expect_match in expected_outputs.items():
            found = re.search(pattern, stream, flags=re.MULTILINE)
            if not bool(found) == expect_match:
                msg = ub.paragraph(
                    f'''
                    Searching for pattern: {pattern!r} in output.
                    Did we expect a match? {expect_match!r}.
                    Did we get a match? {bool(found)!r}.
                    ''')
                raise AssertionError(msg)


def test_kernprof_eager_preimport_bad_module():
    """
    Test for the preservation of the full traceback when an error occurs
    in an auto-generated pre-import module.
    """
    bad_module = '''raise Exception('Boo')'''
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        tmpdir = enter(tempfile.TemporaryDirectory())
        temp_dpath = ub.Path(tmpdir)
        (temp_dpath / 'my_bad_module.py').write_text(bad_module)
        enter(ub.ChDir(tmpdir))
        python_path = os.environ.get('PYTHONPATH', '')
        if python_path:
            python_path = f'{python_path}:{tmpdir}'
        else:
            python_path = tmpdir
        proc = ub.cmd(['kernprof', '-l',
                       # Add an eager pre-import target
                       '-pmy_bad_module', '-c', 'print(1)'],
                      env={**os.environ, 'PYTHONPATH': python_path})
    # Check that the traceback is preserved
    print(proc.stdout)
    print(proc.stderr, file=sys.stderr)
    assert proc.returncode
    assert 'import my_bad_module' in proc.stderr
    assert bad_module in proc.stderr
    # Check that the generated tempfiles are wiped
    reverse_iter_lines = iter(reversed(proc.stderr.splitlines()))
    next(line for line in reverse_iter_lines if 'import my_bad_module' in line)
    tb_header = next(reverse_iter_lines).strip()
    match = re.match('File ([\'"])(.+)\\1, line [0-9]+, in .*', tb_header)
    assert match
    tmp_mod = match.group(2)
    assert not os.path.exists(tmp_mod)
    assert not os.path.exists(os.path.dirname(tmp_mod))


@pytest.mark.parametrize('stdin', [True, False])
def test_kernprof_bad_temp_script(stdin):
    """
    Test for the preservation of the full traceback when an error occurs
    in a temporary script supplied via `kernprof -c` or `kernprof -`.
    """
    bad_script = '''1 / 0'''
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        enter(ub.ChDir(enter(tempfile.TemporaryDirectory())))
        if stdin:
            proc = subprocess.run(
                ['kernprof', '-'],
                input=bad_script, capture_output=True, text=True)
        else:
            proc = subprocess.run(['kernprof', '-c', bad_script],
                                  capture_output=True, text=True)
    # Check that the traceback is preserved
    print(proc.stdout)
    print(proc.stderr, file=sys.stderr)
    assert proc.returncode
    assert '1 / 0' in proc.stderr
    assert 'ZeroDivisionError' in proc.stderr
    # Check that the generated tempfiles are wiped
    reverse_iter_lines = iter(reversed(proc.stderr.splitlines()))
    next(line for line in reverse_iter_lines if '1 / 0' in line)
    tb_header = next(reverse_iter_lines).strip()
    match = re.match('File ([\'"])(.+)\\1, line [0-9]+, in .*', tb_header)
    assert match
    tmp_script = match.group(2)
    assert not os.path.exists(tmp_script)
    assert not os.path.exists(os.path.dirname(tmp_script))


@pytest.mark.parametrize('debug', [True, False])
def test_bad_prof_mod_target(debug):
    """
    Test the handling of bad paths in `--prof-mod` targets.
    """
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        enter(ub.ChDir(enter(tempfile.TemporaryDirectory())))
        proc = ub.cmd(['kernprof', '-l', '-p', './nonexistent.py',
                       '-c', 'print("Output: foo")'],
                      env={**os.environ, 'LINE_PROFILER_DEBUG': str(debug)})
        print(proc.stdout)
        print(proc.stderr, file=sys.stderr)
        proc.check_returncode()
        assert os.listdir()  # Profile results
    assert 'Output: foo' in proc.stdout
    assert re.search(r"1 .* target .*: \['\./nonexistent\.py'\]", proc.stderr)


@pytest.mark.parametrize('builtin', [True, False])
@pytest.mark.parametrize('module', [True, False])
def test_call_with_diagnostics(module, builtin):
    """
    Test the output of call signatures in debug messages.
    """
    to_run = ['-m', 'calendar'] if module else ['-c', 'print("Output: foo")']
    with contextlib.ExitStack() as stack:
        enter = stack.enter_context
        enter(ub.ChDir(enter(tempfile.TemporaryDirectory())))
        cmd = ['kernprof']
        if builtin:
            cmd += ['-b']
        proc = ub.cmd(cmd + to_run,
                      env={**os.environ, 'LINE_PROFILER_DEBUG': 'true'})
        print(proc.stdout)
        print(proc.stderr, file=sys.stderr)
        proc.check_returncode()
        assert os.listdir()  # Profile results
    has_runctx_call = re.search(
        r'Calling: .+\.runctx\(.+\)', proc.stdout, flags=re.DOTALL)
    has_execfile_call = re.search(
        r'execfile\(.+\)', proc.stdout, flags=re.DOTALL)
    assert bool(has_runctx_call) == (not builtin)
    assert bool(has_execfile_call) == (not module)


class TestKernprof(unittest.TestCase):

    def test_enable_disable(self):
        profile = ContextualProfile()
        self.assertEqual(profile.enable_count, 0)
        profile.enable_by_count()
        self.assertEqual(profile.enable_count, 1)
        profile.enable_by_count()
        self.assertEqual(profile.enable_count, 2)
        profile.disable_by_count()
        self.assertEqual(profile.enable_count, 1)
        profile.disable_by_count()
        self.assertEqual(profile.enable_count, 0)
        profile.disable_by_count()
        self.assertEqual(profile.enable_count, 0)

        with profile:
            self.assertEqual(profile.enable_count, 1)
            with profile:
                self.assertEqual(profile.enable_count, 2)
            self.assertEqual(profile.enable_count, 1)
        self.assertEqual(profile.enable_count, 0)

        with self.assertRaises(RuntimeError):
            self.assertEqual(profile.enable_count, 0)
            with profile:
                self.assertEqual(profile.enable_count, 1)
                raise RuntimeError()
        self.assertEqual(profile.enable_count, 0)

    def test_function_decorator(self):
        profile = ContextualProfile()
        f_wrapped = profile(f)
        self.assertEqual(f_wrapped.__name__, f.__name__)
        self.assertEqual(f_wrapped.__doc__, f.__doc__)

        self.assertEqual(profile.enable_count, 0)
        value = f_wrapped(10)
        self.assertEqual(profile.enable_count, 0)
        self.assertEqual(value, f(10))

    def test_gen_decorator(self):
        profile = ContextualProfile()
        g_wrapped = profile(g)
        self.assertEqual(g_wrapped.__name__, g.__name__)
        self.assertEqual(g_wrapped.__doc__, g.__doc__)

        self.assertEqual(profile.enable_count, 0)
        i = g_wrapped(10)
        self.assertEqual(profile.enable_count, 0)
        self.assertEqual(next(i), 20)
        self.assertEqual(profile.enable_count, 0)
        self.assertEqual(i.send(30), 50)
        self.assertEqual(profile.enable_count, 0)

        with self.assertRaises((StopIteration, RuntimeError)):
            next(i)
        self.assertEqual(profile.enable_count, 0)


if __name__ == '__main__':
    """
    CommandLine:
        python ./tests/test_kernprof.py
    """
    unittest.main()



"""
Read and resolve user-supplied TOML files and combine them with the
default to generate configurations.
"""
import copy
import dataclasses
import functools
import importlib.resources
import itertools
import os
import pathlib
try:
    import tomllib
except ImportError:  # Python < 3.11
    import tomli as tomllib  # type: ignore[no-redef] # noqa: F811
from collections.abc import Mapping
from typing import Dict, List, Any


__all__ = ['ConfigSource']

NAMESPACE = 'tool', 'line_profiler'
TARGETS = 'line_profiler.toml', 'pyproject.toml'
ENV_VAR = 'LINE_PROFILER_RC'

_DEFAULTS = None


@dataclasses.dataclass
class ConfigSource:
    """
    Object encapsulating the config dict and the source whence it is
    read from.

    Attributes:
        conf_dict (dict[str, Any])
            The combination of the ``tool.line_profiler`` tables of the
            provided/looked-up config file (if any) and the default as a
            dictionary.
        path (pathlib.Path)
            Absolute path to the config file whence the config options
            are loaded.
        subtable (list[str])
            Sequence of table headers under which in
            :py:attr:`~.ConfigSource.path`
            :py:attr:`~.ConfigSource.conf_dict` can be found.
    """
    conf_dict: Dict[str, Any]
    path: pathlib.Path
    subtable: List[str]

    def copy(self):
        """
        Returns:
            Copy of the object.
        """
        return type(self)(
            copy.deepcopy(self.conf_dict), self.path, self.subtable.copy())

    def get_subconfig(self, *headers, allow_absence=False, copy=False):
        """
        Arguments:
            headers (str):
                Table headers.
            allow_absence (bool):
                If true, allow for the keys to be absent (and return an
                instance with an empty
                :py:attr:`~.ConfigSource.conf_dict`);
                otherwise, raise a :py:class:`KeyError`.
            copy (bool):
                If true, create a (deep) copy of the subtable in
                ``self`` for the new instance's
                :py:attr:`~.ConfigSource.conf_dict`;
                otherwise, just refer to the existing subtable.

        Returns:
            New instance which consists of the required subtable of the
            existing one.

        Example:
            >>> default = ConfigSource.from_default()
            >>> display_widths = default.get_subconfig(
            ...     'show', 'column_widths')
            >>> assert display_widths.path == default.path
            >>> assert (display_widths.subtable
            ...         == default.subtable + ['show', 'column_widths'])
            >>> assert (display_widths.conf_dict
            ...         is default.conf_dict['show']['column_widths'])
        """
        new_dict = get_subtable(
            self.conf_dict, headers, allow_absence=allow_absence)
        new_subtable = [*self.subtable, *headers]
        return type(self)(new_dict, self.path, new_subtable)

    @classmethod
    def from_default(cls, *, copy=True):
        """
        Get the default TOML configuration that ships with the package.

        Arguments:
            copy (bool):
                Whether to make a copy.

        Returns:
            New instance if ``copy`` is true, the global default
            instance otherwise.
        """
        # Note: `importlib.resources.path()` is deprecated on 3.11 and
        # legacy patch versions of 3.12, and only later un-deprecated
        # on 3.13 onwards. So we use the newer APIs where available.
        # (See the discussions on GitHub issue #405)
        ir = importlib.resources
        try:
            ir_files, ir_as_file = ir.files, ir.as_file
        except AttributeError:  # Python < 3.9
            find_file = ir.path
        else:
            def find_file(anc, *chunks):
                return ir_as_file(ir_files(anc).joinpath(*chunks))

        global _DEFAULTS
        if _DEFAULTS is None:
            package = __spec__.name.rpartition('.')[0]
            with find_file(package + '.rc', 'line_profiler.toml') as path:
                conf_dict, source = find_and_read_config_file(config=path)
            conf_dict = get_subtable(conf_dict, NAMESPACE, allow_absence=False)
            _DEFAULTS = cls(conf_dict, source, list(NAMESPACE))
        if not copy:
            return _DEFAULTS
        return _DEFAULTS.copy()

    @classmethod
    def from_config(cls, config=None, *, read_env=True):
        """
        Create an instance by loading from a config file.

        Arguments:
            config (str | os.PathLike[str] | bool | None):
                Optional path to a specific TOML file;
                if a (string) path, try to read from that file;
                if :py:data:`None` or :py:data:`True`, look up and
                resolve to the correct file;
                if :py:data:`False`, just return a copy of the default
                config (see :py:meth:`~.ConfigSource.from_default`).
            read_env (bool):
                How to look up the config file if not provided (i.e.
                ``config = None`` or equivalently :py:data:`True`):

                :py:data:`True`
                    Try to read the environment variable
                    :envvar:`!LINE_PROFILER_RC` as the path to a config
                    file;
                    if that fails, fall back to the default configuation
                    (see :py:meth:`~.ConfigSource.from_default`).

                :py:data:`False`
                    Use path-based lookup (see Note) to resolve to a
                    config file.

        Returns:
            New instance

        Note:
            * For the config TOML file, it is required that each of the
              following keys either is absent or maps to a table:

              * ``tool`` and ``tool.line_profiler``
              * ``tool.line_profiler.kernprof``, ``.cli``, ``.setup``,
                ``.write``, and ``.show``
              * ``tool.line_profiler.show.column_widths``

              If this is not the case:

              * If ``config`` is provided, a :py:class:`ValueError` is
                raised.
              * Otherwise, the looked-up file is considered invalid and
                ignored.
            * When performing path-based lookup:

              * The current directory is checked first to see if it has
                a valid, readable TOML file named ``line_profiler.toml``.
              * If not, check if there is a valid, readable TOML file
                named ``pyproject.toml``.
              * If not, check the parent directory, and so on.
              * If we reached the file-system root without finding a
                valid, readable TOML file, fall back to the default
                configuration (see
                :py:meth:`~.ConfigSource.from_default`).
        """
        def merge(template, supplied):
            if not (isinstance(template, dict) and isinstance(supplied, dict)):
                return supplied
            result = {}
            for key, default in template.items():
                if key in supplied:
                    result[key] = merge(default, supplied[key])
                else:
                    result[key] = default
            return result

        default_instance = cls.from_default()
        if config in (True, False):
            if config:
                config = None
            else:
                return default_instance
        if config is not None:
            # Promote to `Path` (and catch type errors) early
            config = pathlib.Path(config)
        if read_env:
            get_conf = functools.partial(find_and_read_config_file,
                                         config=config)
        else:  # Shield the lookup from the environment
            get_conf = functools.partial(find_and_read_config_file,
                                         config=config, env_var=None)
        try:
            content, source = get_conf()
        except TypeError:  # Got `None`
            if config:
                if os.path.exists(config):
                    Error = ValueError
                else:
                    Error = FileNotFoundError
                raise Error(
                    f'Cannot load configurations from {config!r}') from None
            return default_instance
        conf = {}
        try:
            for header in get_headers(default_instance.conf_dict):
                # Get the top-level subtable
                key, *subheader = header
                subtable = get_subtable(content, [*NAMESPACE, key])
                # Check the existence of nested subtables (if any)
                get_subtable(subtable, subheader)
                # If it looks OK, remember the top-level subtable
                conf.setdefault(key, subtable)
        except (TypeError, AttributeError):
            if config is None:
                # No file explicitly provided and the looked-up file is
                # invalid, just fall back to the default configs
                return default_instance
            else:
                # The explicitly provided config file is invalid, raise
                # an error
                all_headers = {'tool', 'tool.line_profiler'}
                all_headers.update(
                    '.'.join(('tool.line_profiler', *header))
                    for header in get_headers(default_instance.conf_dict,
                                              include_implied=True))
                raise ValueError(
                    f'config = {config!r}: expected each of these keys to '
                    'either be nonexistent or map to a table: '
                    f'{sorted(all_headers)!r}') from None
        # Filter the content of `conf` down to just the key-value pairs
        # pairs present in the default configs
        return cls(
            merge(default_instance.conf_dict, conf), source, list(NAMESPACE))


def find_and_read_config_file(
        *, config=None, env_var=ENV_VAR, targets=TARGETS):
    """
    Arguments:
        config (str | os.PathLike[str] | None):
            Optional path to a specific TOML file;
            if provided, skip lookup and just try to read from that file
        env_var (str | None):
            Name of the of the environment variable containing the path
            to a TOML file;
            if true-y and if ``config`` isn't provided, skip lookup and
            just try to read from that file
        targets (Sequence[str | os.PathLike[str]]):
            Filenames among which TOML files are looked up (if neither
            ``config`` or ``env_var`` is given)

    Returns:
        If the provided/looked-up file is readable and is valid TOML:
            tuple[dict, Path]: content, path
                * ``content``: parsed content of the file as a
                  dictionary
                * ``path``: absolute path to the file
        Otherwise
            None
    """
    def iter_configs(dir_path):
        for dpath in itertools.chain((dir_path,), dir_path.parents):
            for target in targets:
                cfg = dpath / target
                try:
                    if cfg.is_file():
                        yield cfg
                except OSError:  # E.g. permission errors
                    pass

    if config:
        configs = pathlib.Path(config).absolute(),
    elif env_var and os.environ.get(env_var):
        configs = pathlib.Path(os.environ[env_var]).absolute(),
    else:
        pwd = pathlib.Path.cwd().absolute()
        configs = iter_configs(pwd)
    for config in configs:
        try:
            with config.open(mode='rb') as fobj:
                return tomllib.load(fobj), config
        except (OSError, tomllib.TOMLDecodeError):
            pass
    return None


def get_subtable(table, keys, *, allow_absence=True):
    """
    Arguments:
        table (Mapping):
            (Nested) Mapping.
        keys (Sequence):
            Sequence of keys for item access on ``table`` and its
            descendant tables.
        allow_absence (bool):
            If true, allow for the keys to be absent;
            otherwise, raise a :py:class:`KeyError`.

    Returns:
        Mapping: subtable

    Example:

        >>> table = {'a': 1, 'b': {'c': 2, 'd': 3, 'e': {}}}
        >>> assert get_subtable(table, []) == table
        >>> assert get_subtable(table, ['b']) == table['b']
        >>> assert get_subtable(table, ['b', 'e']) == table['b']['e']
        >>> assert get_subtable(table, ['c']) == {}
        >>> get_subtable(table, ['c'], allow_absence=False)
        Traceback (most recent call last):
          ...
        KeyError: 'c'
        >>> get_subtable(  # doctest: +ELLIPSIS, +NORMALIZE_WHITESPACE
        ...     table, ['a'])
        Traceback (most recent call last):
          ...
        TypeError: table = ..., keys = ['a']:
        expected result to be a mapping, got a `int` (1)
    """
    subtable = table
    for key in keys:
        if allow_absence:
            subtable = subtable.get(key, {})
        else:
            subtable = subtable[key]
    if not isinstance(subtable, Mapping):
        raise TypeError(f'table = {table!r}, keys = {list(keys)!r}: '
                        'expected result to be a mapping, got a '
                        f'`{type(subtable).__name__}` ({subtable!r})')
    return subtable


def get_headers(table, *, include_implied=False):
    """
    Arguments:
        table (Mapping):
            (Nested) Mapping.
        include_implied (bool):
            if false and if a subtable has other subtables, only the
            terminal key sequences are returned (that is to say, if
            ``table['a']['b']`` is a subtable, then ``('a',)`` is only
            in ``headers`` if ``include_implied`` is true.

    Returns:
        set[tuple]: headers
            Key sequences corresponding to the subtables of ``table``.

    Example:

        >>> table = {'a': 1,
        ...          'b': {'c': 2, 'd': 3, 'e': {}, 'f': {'g': 4}},
        ...          'h': {'i': 5}}
        >>> assert get_headers(table) == {
        ...     ('b', 'e'), ('b', 'f'), ('h',)}
        >>> assert get_headers(table, include_implied=True) == {
        ...     ('b',), ('b', 'e'), ('b', 'f'), ('h',)}
        >>> assert get_headers({}) == set()
        >>> assert get_headers({'a': 1, 'b': 2}) == set()
    """
    results = set()
    for key, value in table.items():
        if not isinstance(value, Mapping):
            continue
        subheaders = get_headers(value, include_implied=include_implied)
        if subheaders:
            results.update((key,) + header for header in subheaders)
        if include_implied or not subheaders:
            results.add((key,))
    return results



"""
Shared utilities between the :command:`python -m line_profiler` and
:command:`kernprof` CLI tools.
"""
import argparse
import pathlib
from os import PathLike
from typing import Protocol, Sequence, Tuple, TypeVar

from line_profiler.toml_config import ConfigSource


P_con = TypeVar('P_con', bound='ParserLike', contravariant=True)
A_co = TypeVar('A_co', bound='ActionLike', covariant=True)


class ActionLike(Protocol[P_con]):
    def __call__(self, parser: P_con,
                 namespace: argparse.Namespace,
                 values: str | Sequence | None,
                 option_string: str | None = None) -> None:
        ...

    def format_usage(self) -> str:
        ...


class ParserLike(Protocol[A_co]):
    def add_argument(self, arg: str, /, *args: str, **kwargs) -> A_co:
        ...

    @property
    def prefix_chars(self) -> str:
        ...


def add_argument(parser_like: ParserLike[A_co], arg: str, /, *args: str,
                 hide_complementary_options: bool = True, **kwargs) -> A_co:
    ...


def get_cli_config(subtable: str, /, *args, **kwargs) -> ConfigSource:
    ...


def get_python_executable() -> str:
    ...


def positive_float(value: str) -> float:
    ...


def boolean(value: str, *,
            fallback: bool | None = None, invert: bool = False) -> bool:
    ...


def short_string_path(path: str | PathLike[str]) -> str:
    ...
